import torch.nn as nn
import torch
import numpy as np
from torch_geometric.nn import GlobalAttention, global_mean_pool
from ..data.data_tools import edges_geom_causal
from torch_sparse import SparseTensor
from torch_geometric.data import Batch
from torch.nn.utils import weight_norm

## Basic perceptron


class ExpLayer(nn.Module):
    def forward(self, x):
        return torch.exp(x)


class BoundSigmoid(nn.Module):
    def __init__(self, out_range):
        super(BoundSigmoid, self).__init__()
        self.out_range = out_range
        assert self.out_range[0] < self.out_range[1]

    def forward(self, x):
        return (
            torch.sigmoid(x) * (self.out_range[1] - self.out_range[0])
            + self.out_range[0]
        )


class ResidualNet(nn.Module):
    def __init__(self, mod):
        super(ResidualNet, self).__init__()
        self.mod = mod

    def forward(self, x):
        M = self.mod(x)
        return 0.1 * M + x[:, : M.shape[1]]


def MLP(
    channels,
    activation="leaky",
    last_activation="identity",
    use_batch_norm=False,  # used to be True. this is very unstable with True...
    bias=True,
    dropout=0.0,
    out_range=None,
    residual=False,
    use_weight_norm=True,  # False,
):

    if out_range is not None:
        last_activation = BoundSigmoid(out_range)
    elif last_activation == "identity":
        last_activation = nn.Identity()
    elif last_activation == "exponential":
        last_activation = ExpLayer()

    if activation == "leaky":
        activations = [nn.LeakyReLU(0.2) for i in range(1, len(channels) - 1)] + [
            last_activation
        ]
    elif activation == "ELU":
        activations = [nn.ELU() for i in range(1, len(channels) - 1)] + [
            last_activation
        ]
    else:
        activations = [nn.ReLU() for i in range(1, len(channels) - 1)] + [
            last_activation
        ]

    norm_function = lambda lin: lin
    if use_weight_norm:
        norm_function = lambda lin: weight_norm(lin)

    sequential = nn.Sequential(
        *[
            nn.Sequential(
                nn.Dropout(dropout) if dropout > 0.0 else nn.Identity(),
                norm_function(nn.Linear(channels[i - 1], channels[i], bias=bias)),
                nn.BatchNorm1d(channels[i]) if use_batch_norm else nn.Identity(),
                activations[i - 1],
            )
            for i in range(1, len(channels))
        ]
    )

    if not residual:
        return sequential
    else:
        return ResidualNet(sequential)


def batch_from_positions(pos, N, L, D, degree):
    x_pos = torch.reshape(pos, (N * L, D))
    # assert torch.equal(x_pos[L : (2 * L)], pos[1])
    batch = torch.repeat_interleave(torch.arange(N, device=pos.device), L)
    # edge_index = x.get_edges(steps[0], {"edge_method": "geom_causal"})
    row, col = edges_geom_causal(L, degree)
    row = torch.from_numpy(row).to(x_pos.device)
    col = torch.from_numpy(col).to(x_pos.device)
    edge_index = torch.stack((row, col), dim=0).long()

    N_edges = edge_index.shape[1]
    edge_index = edge_index.repeat((1, N))
    shift = (
        torch.repeat_interleave(torch.arange(N, device=x_pos.device).long(), N_edges)
        * L
    )
    edge_index += torch.stack((shift, shift), dim=0)
    # adj_t = SparseTensor(col=edge_index[0], row=edge_index[1])

    x_pred = Batch(batch=batch, pos=x_pos, edge_index=edge_index)

    return x_pred


def batch_from_sub_batches(sub_batches):
    pos_vectors = []
    edge_index_offset = 0
    batch_offset = 0
    batch_vectors = []
    nodes_offset = 0
    edge_indices = []

    assert "edge_index" in sub_batches[0].keys

    other_keys = list(
        set(sub_batches[0].keys) - set(["pos", "ptr", "edge_index", "batch"])
    )
    other_keys_dict = {}

    for k in other_keys:
        other_keys_dict[k] = []

    for batch in sub_batches:
        B = batch.batch + batch_offset
        batch_vectors.append(B)
        batch_offset = torch.max(B) + 1

        pos_vectors.append(batch.pos)

        e = batch.edge_index + nodes_offset
        edge_indices.append(e)
        nodes_offset += batch.pos.shape[0]

        for k in other_keys:
            other_keys_dict[k].append(batch[k])

    edge_index = torch.cat(edge_indices, dim=1)
    pos = torch.cat(pos_vectors, dim=0)
    batch = torch.cat(batch_vectors, dim=0)

    assert torch.max(edge_index) == pos.shape[0] - 1
    assert torch.min(edge_index[1:, 1] - edge_index[:-1, 1]) >= 0

    adj_t = SparseTensor(col=edge_index[0], row=edge_index[1])
    new_batch = Batch(batch=batch, pos=pos, adj_t=adj_t)

    for k in other_keys:
        try:
            new_batch[k] = torch.cat(other_keys_dict[k], dim=0)
        except Exception as e:
            print("key = %s" % k)
            raise e

    return new_batch


def generate_batch_like(
    x,
    T_values,
    alpha_range,
    tau_range,
    generator,
    degree,
    simulate_tau=False,
    simulate_diffusion=True,
):
    batches = []
    BS = torch.max(x.batch) + 1
    # print("BS = %d" % BS)
    SBS_min = BS // len(T_values)
    # print("SBS = %s" % SBS)

    # if not x.pos.is_cuda:
    #    x.pos = x.pos.cuda()

    for T in T_values:
        SBS = int(SBS_min)
        # print("Generate %d trajs of length %d" % (SBS, T))
        # print("T = %d, generating %d trajs" % (T, SBS))
        # ALPHA
        alpha = (
            torch.rand(SBS, device=x.pos.device) * (alpha_range[1] - alpha_range[0])
            + alpha_range[0]
        )
        # print("alpha = ", alpha)

        # TAU if needed
        if simulate_tau:
            log_tau = torch.rand(SBS, device=x.pos.device) * (
                np.log10(tau_range[1]) - np.log10(tau_range[0])
            ) + np.log10(tau_range[0])
        else:
            log_tau = torch.ones_like(alpha) * np.log10(T) + 1
        tau = torch.floor(torch.pow(10.0, log_tau))
        log_tau = torch.log10(tau)

        # print("tau = ", tau)
        # Make sure that tau is not larger than T
        # tau = torch.where(tau > T, T, tau)

        # DIFFUSION
        if simulate_diffusion == True:
            log_diffusion = torch.rand(SBS, device=x.pos.device) * 4 - 2
        else:
            log_diffusion = torch.ones_like(alpha) * 0.0
        diffusion = torch.pow(10.0, log_diffusion)

        pos = generator(alpha, tau, diffusion, T)
        # print("Generating T = %d" % T)
        # print(pos.shape)

        x = batch_from_positions(
            pos,
            N=SBS,
            L=T,
            D=generator.dim,
            degree=degree,
        )
        x.alpha = alpha.view(-1, 1)
        x.log_tau = log_tau.view(-1, 1)
        x.log_diffusion = log_diffusion.view(-1, 1)
        x.length = torch.ones_like(x.alpha) * T
        assert x.batch.shape[0] == SBS * T
        batches.append(x)

    # print("num of sub_batches %d " % len(batches))

    x = batch_from_sub_batches(batches)

    return x


class AlphaPredictor(nn.Module):
    """
    The only interest of this class is to add an offset and to restrict the output to a plausible interval
    """

    def __init__(
        self,
        p=0.0,
        input_dim=128,
        mlp_size=[128, 128, 64, 16],
    ):

        super(AlphaPredictor, self).__init__()
        MLP_size = [input_dim] + mlp_size + [1]
        self.mlp = MLP(MLP_size, dropout=p)

    def forward(self, x):
        residual = self.mlp(x)
        return 1.0 + 0.99 * torch.nn.Tanh()(residual)



import torch
import numpy as np
from sklearn.metrics import f1_score
from ..data.datamodule import EMPTY_FIELD_VALUE
import torch.nn as nn
import torch

## Losses


def L2_loss(out, target, w):
    # w : whether each sample has to be included
    # for instance, OU don't have to be included in alpha,
    # while LW don't have to be included in theta
    mean_dims = torch.mean(torch.pow(out[w] - target[w], 2), dim=1)
    # concerned = torch.masked_select(mean_dims,w)
    return torch.mean(mean_dims)


def Category_loss(out, target, w):
    # w : whether each sample has to be included
    # return nn.CrossEntropyLoss()(torch.index_select(out,0,w), torch.index_select(target,0,w))
    l = nn.CrossEntropyLoss()(
        out[w],
        target[w].view(
            -1,
        ),
    )
    return l


## Metrics for training
def is_concerned(target):
    """Returns 1 if the sample should be included in the loss (based on the value of the target)

    Args:
        target (_type_): _description_

    Returns:
        _torch.Tensor_: tensor of booleans
    """
    if len(target.shape) == 1:
        return ~torch.isnan(target)
    else:
        return torch.eq(torch.sum(1 * torch.isnan(target), dim=1), 0)
    # return torch.eq(torch.sum(1 * torch.eq(target, EMPTY_FIELD_VALUE), dim=1), 0)



from .diverse import *
from torch_geometric.nn import MessagePassing
from torch_sparse import matmul


class MinimalJumpsConv(MessagePassing):
    def __init__(
        self,
        out_channels,
        x_dim,
        dropout=0.0,
        aggr="mean",
        moments=[1],
        f_inner_width=[128, 64],
        **kwargs
    ):
        super(MinimalJumpsConv, self).__init__(
            aggr=aggr, **kwargs
        )  # , flow="target_to_source"
        self.out_channels = out_channels
        self.p = dropout
        self.moments = moments
        M = len(moments) + 1
        self.bn_x = nn.BatchNorm1d(x_dim)

        MLP_size = [M * x_dim] + f_inner_width + [out_channels]
        if M > 1:
            self.f = nn.Sequential(
                nn.BatchNorm1d(M * x_dim), MLP(MLP_size, dropout=dropout)
            )  # last_activation=nn.Tanh()))
        else:
            self.f = MLP(MLP_size, dropout=dropout)

        nparams = 0
        for n, p in self.named_parameters():
            np_ = np.product(np.array([s for s in p.shape]))
            nparams += np_
        # print("f size = ", MLP_size)
        # print(
        #    "Convolution has %d parameters. Input dim is %d, output is %d"
        #    % (nparams, x_dim, out_channels)
        # )

    def forward(self, x, edge_index):  # removed edge_attr

        x = self.bn_x(x)

        neighbors_message = self.propagate(x=x, edge_index=edge_index)

        neighbors_message = torch.cat(
            [x] + [torch.pow(neighbors_message, m) for m in self.moments], dim=1
        )
        result = self.f(neighbors_message)

        return result

    def message_and_aggregate(self, adj_t, x):
        # print("message and aggregate")
        return matmul(adj_t, x, reduce=self.aggr)



from torch._C import ThroughputBenchmark
from .diverse import *
from .minimal_conv import *
from torch_geometric.nn import NNConv, global_mean_pool, GCNConv
from torch_sparse import SparseTensor


class TrajsEncoder2(nn.Module):
    def __init__(
        self,
        traj_dim: int,
        x_dim: int,
        e_dim: int,
        n_c: int = 64,  # Number of convolution kernels
        n_scales: int = 1,
        latent_dim: int = 8,
    ):
        super(TrajsEncoder2, self).__init__()
        e_latent_dim = 16

        self.nodes_MLP = MLP([x_dim, 128, x_dim], use_batch_norm=True)

        self.edges_MLP = MLP([e_dim, 128, e_latent_dim], use_batch_norm=True)

        self.att_conv = GCNConv(
            in_channels=x_dim,
            out_channels=n_c,
            improved=True,
        )

        self.conv_edges = NNConv(
            in_channels=n_c,
            out_channels=n_c,
            nn=MLP([e_latent_dim, e_latent_dim * 2, n_c * n_c]),
            aggr="mean",
        )

        self.last_conv = GCNConv(n_c, n_c, aggr="max", improved=True)

        gate_nn = MLP([3 * n_c, 256, 1])
        self.pooling = GlobalAttention(gate_nn=gate_nn)
        # self.pooling = global_mean_pool

        self.n_scales = n_scales
        self.traj_dim = traj_dim

        self.mlp = MLP(
            [3 * n_c + n_scales + traj_dim, 2 * latent_dim, latent_dim, latent_dim]
        )  # used to be tanh for last_activation

    def forward(self, data):
        adj_t = data.adj_t
        # print(data.adj_t[:, :, 0])
        row, col, edge_attr = adj_t.t().coo()
        sparse_adj_t = SparseTensor(col=col, row=row)
        edge_index = torch.stack([row, col], dim=0)

        x1 = self.att_conv(
            x=self.nodes_MLP(data.x),
            edge_index=sparse_adj_t,
        )

        edges_embedding = self.edges_MLP(edge_attr)

        # adj_t = adj_t.set_value(edges_embedding)

        # x2 = self.conv_edges(x=x1, edge_index=adj_t)
        x2 = self.conv_edges(x=x1, edge_index=edge_index, edge_attr=edges_embedding)
        x2 = torch.tanh(x2)

        x3 = self.last_conv(x=x2, edge_index=sparse_adj_t)

        x = torch.cat([x1, x2, x3], dim=1)
        # x = x1 + x2 + x3

        x = self.pooling(x=x, batch=data.batch)

        # print(x.size())

        if self.n_scales > 0:
            x = torch.cat((x, torch.log(data.scales + 1e-5)), dim=1)
        if self.traj_dim > 0:
            x = torch.cat((x, data.orientation), dim=1)

        # print(x.size())
        out = self.mlp(x)

        return out



from operator import mod
from pyexpat import model
from gratin.models.main_net import MainNet
from gratin.data.datamodule import DataModule
import pytorch_lightning as pl
from gratin.models.utils import get_predictions_of_dl
from gratin.training.callbacks import LatentSpaceSaver, Plotter
import matplotlib.pyplot as plt
from pytorch_lightning.callbacks import (
    EarlyStopping,
    LearningRateMonitor,
    ModelCheckpoint,
    StochasticWeightAveraging,
)
import os
from umap import ParametricUMAP
import tensorflow as tf
import numpy as np
import logging
import torch.cuda

logging.getLogger("pytorch_lightning").setLevel(logging.ERROR)

graph_info = {
    "edges_per_point": 10,
    "clip_trajs": False,
    "scale_types": ["step_sum", "pos_std", "step_std"],
    "log_features": True,
    "data_type": "no_features",  # no features because features are all computed by the model
    "edge_method": "geom_causal",
}


def _get_encoder_structure():
    return tf.keras.Sequential(
        [
            tf.keras.layers.InputLayer(input_shape=(16)),
            tf.keras.layers.Dense(units=256),
            tf.keras.layers.LeakyReLU(alpha=0.1),
            tf.keras.layers.Dense(units=16),
            tf.keras.layers.LeakyReLU(alpha=0.1),
            tf.keras.layers.Dense(units=2),
        ]
    )


def train_model(
    export_path: str,
    time_delta: float,
    max_n_epochs: int = 100,
    num_workers: int = 0,  # number of workers during the training process (should be < # CPUs)
    dim: int = 2,  # Dimension of trajectories. 2 or 3
    log_diffusion_range: tuple = (
        -2.0,
        1.1,
    ),  # log-diffusion is drawn following a truncated centered gaussian in this range
    length_range: tuple = (7, 35),  # length is drawn in a log-uniform way
    noise_range: tuple = (
        0.015,
        0.05,
    ),  # localization uncertainty, in micrometers (one value per trajectory)
):

    if not os.path.exists(export_path):
        os.mkdir(export_path)
    assert os.path.isdir(export_path)

    dl_params = {"batch_size": 128, "num_workers": num_workers}

    ds_params = {
        "dim": dim,  # can be (1, 2 or 3)
        "RW_types": [
            "fBM",
            "LW",
            "sBM",
            "OU",
            "CTRW",
        ],  # Types of random walks used during training
        "time_delta": time_delta,
        "logdiffusion_range": log_diffusion_range,
        "length_range": length_range,
        "noise_range": noise_range,
        "N": int(1e5),
    }

    model = MainNet(
        # tasks=["alpha", "model"],
        n_c=32,
        latent_dim=16,
        lr=1e-3,
        dim=2,
        RW_types=ds_params["RW_types"],
        scale_types=["step_std", "pos_std"],
    )

    pl.seed_everything(1)
    dm = DataModule(ds_params=ds_params, dl_params=dl_params, graph_info=graph_info)
    dm.setup(plot=False)

    ES = EarlyStopping(
        monitor="train_loss",  # We do not care about overfitting, train loss is more stable than validation
        min_delta=0.001,
        patience=5,
        verbose=True,
        mode="min",
        strict=True,
    )
    LRM = LearningRateMonitor("epoch")
    CKPT = ModelCheckpoint(
        dirpath=export_path,
        filename="model",
        monitor="train_loss",
        verbose=True,
        mode="min",
    )

    tb_logger = pl.loggers.TensorBoardLogger(
        save_dir=os.path.join(export_path, "tb_logs"),
        default_hp_metric=False,
        name="model",
        flush_secs=60,
    )

    trainer = pl.Trainer(
        auto_select_gpus=torch.cuda.is_available(),
        gpus=1 * torch.cuda.is_available(),
        gradient_clip_val=10.0,
        reload_dataloaders_every_n_epochs=1,
        callbacks=[
            ES,
            LRM,
            CKPT,
            Plotter(),
            # LatentSpaceSaver(),
            # StochasticWeightAveraging(swa_epoch_start=10),
        ],
        log_every_n_steps=50,
        max_epochs=max_n_epochs,
        detect_anomaly=True,
        track_grad_norm=2,
        logger=tb_logger,
    )

    trainer.fit(model=model, datamodule=dm)

    results = trainer.test(model, dm.test_dataloader())
    print(results)

    model.eval()
    output, target, h = get_predictions_of_dl(
        model, dm.test_dataloader(), latent_samples=int(3e4)
    )

    umap_params = {
        "min_dist": 0.5,
        "n_components": 2,
        "verbose": True,
        "n_neighbors": 200,
        "metric": "cosine",
    }

    encoder = _get_encoder_structure()

    U = ParametricUMAP(**umap_params, encoder=encoder)
    U.fit(h)

    encoder.save(os.path.join(export_path, "umap"))
    X = np.array(encoder(h))

    alpha = output["alpha"]
    alpha_true = target["alpha"]
    true_model = target["model"]

    plt.figure()
    cond = alpha_true > 0
    cm = plt.scatter(X[cond, 0], X[cond, 1], c=target["log_diffusion"][cond], s=0.1)
    plt.colorbar(cm, label="$\\log_{10}(D)$")
    plt.tight_layout()

    plt.figure()
    cond = alpha_true > 0
    cm = plt.scatter(X[cond, 0], X[cond, 1], c=alpha_true[cond], s=0.1)
    plt.colorbar(cm, label="$\\alpha$")
    plt.tight_layout()

    plt.figure()
    for m in np.unique(true_model):
        plt.scatter(
            X[true_model == m, 0],
            X[true_model == m, 1],
            s=1,
            label=ds_params["RW_types"][m],
        )
    plt.legend(markerscale=10)
    plt.tight_layout()

    plt.figure()
    mean_err = [
        np.mean(
            np.abs(
                alpha[(target["length"] == L) & (alpha_true > 0)]
                - alpha_true[(target["length"] == L) & (alpha_true > 0)]
            )
        )
        for L in np.unique(target["length"])
    ]
    length = np.unique(target["length"])
    plt.plot(length, mean_err)
    plt.ylabel("MAE($\\alpha$)")
    plt.xlabel("Trajectory length")
    plt.tight_layout()

    return model, encoder


def load_model(export_path: str):
    model = MainNet.load_from_checkpoint(os.path.join(export_path, "model.ckpt"))
    model.eval()
    model.freeze()

    encoder = tf.keras.models.load_model(os.path.join(export_path, "umap"))
    return model, encoder


def plot_demo(
    model: MainNet,
    encoder: tf.keras.Model,
    num_workers: int = 0,
    dim: int = 2,
    time_delta: float = 0.03,
    log_diffusion_range: tuple = (-2.0, 1.1),
    length_range: tuple = (7, 35),
    noise_range: tuple = (0.015, 0.05),
):

    dl_params = {"batch_size": 128, "num_workers": num_workers}

    ds_params = {
        "dim": dim,  # can be (1, 2 or 3)
        "RW_types": [
            "fBM",
            "LW",
            "sBM",
            "OU",
            "CTRW",
        ],  # Types of random walks used during training
        "time_delta": time_delta,
        "logdiffusion_range": log_diffusion_range,
        "length_range": length_range,
        "noise_range": noise_range,
        "N": int(1e5),
    }

    pl.seed_everything(1)
    dm = DataModule(ds_params=ds_params, dl_params=dl_params, graph_info=graph_info)
    dm.setup(plot=False)

    trainer = pl.Trainer(
        auto_select_gpus=torch.cuda.is_available(), gpus=1 * torch.cuda.is_available()
    )

    trainer.test(model, dm.test_dataloader())

    output, target, h = get_predictions_of_dl(
        model, dm.test_dataloader(), latent_samples=int(3e4)
    )

    X = np.array(encoder(h))

    alpha = output["alpha"]
    alpha_true = target["alpha"]
    true_model = target["model"]

    plt.figure()
    cond = alpha_true > 0
    cm = plt.scatter(X[cond, 0], X[cond, 1], c=target["log_diffusion"][cond], s=0.1)
    plt.colorbar(cm, label="$\\log_{10}(D)$")
    plt.tight_layout()

    plt.figure()
    cond = alpha_true > 0
    cm = plt.scatter(X[cond, 0], X[cond, 1], c=alpha_true[cond], s=0.1)
    plt.colorbar(cm, label="$\\alpha$")
    plt.tight_layout()

    plt.figure()
    for m in np.unique(true_model):
        plt.scatter(
            X[true_model == m, 0],
            X[true_model == m, 1],
            s=1,
            label=ds_params["RW_types"][m],
        )
    plt.legend(markerscale=10)
    plt.tight_layout()

    plt.figure()
    mean_err = [
        np.mean(
            np.abs(
                alpha[(target["length"] == L) & (alpha_true > 0)]
                - alpha_true[(target["length"] == L) & (alpha_true > 0)]
            )
        )
        for L in np.unique(target["length"])
    ]
    length = np.unique(target["length"])
    plt.plot(length, mean_err)
    plt.ylabel("MAE($\\alpha$)")
    plt.xlabel("Trajectory length")
    plt.tight_layout()


def get_predictions(model, encoder, trajectories):
    assert len(trajectories) > 0, "Empty list of trajectories"
    from gratin.models.utils import get_predictions_of_dl
    from gratin.data.dataset import ExpTrajDataSet
    from gratin.standard import graph_info
    from torch_geometric.loader import DataLoader
    import pandas as pd

    indices = np.arange(len(trajectories))
    exp_ds = ExpTrajDataSet(
        dim=trajectories[0].shape[1], graph_info=graph_info, trajs=trajectories
    )
    dl_exp = DataLoader(exp_ds, batch_size=128, shuffle=False, drop_last=False)

    outputs, info, h = get_predictions_of_dl(model, dl_exp, latent_samples=len(exp_ds))
    probabilities = np.exp(outputs["model"]) / np.reshape(
        np.exp(outputs["model"]).sum(axis=1), (-1, 1)
    )
    best_model = [
        model.hparams["RW_types"][i] for i in np.argmax(probabilities, axis=1)
    ]
    df = pd.DataFrame(index=indices)
    df["alpha"] = outputs["alpha"]
    df["best_model"] = best_model
    df[["p_%s" % m for m in model.hparams["RW_types"]]] = probabilities
    df[["U_1", "U_2"]] = np.array(encoder(h))
    h_cols = ["h_%d" % (i + 1) for i in range(h.shape[1])]
    df[h_cols] = h

    return df



import numpy as np
import sys
import os


def get_traj_scale(traj):
    diff = traj[1:] - traj[:-1]
    dr_2 = np.sum(diff ** 2, axis=1)
    scale = np.sqrt(np.mean(dr_2))
    if scale == 0:
        return 1.0
    else:
        return scale


def sample_sphere(dim):
    X = np.random.normal(0, 1, dim)
    X /= np.linalg.norm(X, ord=2)
    return X


def get_drift(traj, drift_amplitude):
    dim = traj.shape[1]
    drift_norm = drift_amplitude * get_traj_scale(traj)
    u = sample_sphere(dim)
    drift = drift_norm * u
    return drift, u


def add_drift(traj, drift):
    L, dim = traj.shape
    d_pos = np.repeat(np.reshape(drift, (1, -1)), L, axis=0)
    d_pos = d_pos * np.repeat(np.reshape(np.arange(L), (-1, 1)), dim, axis=1)
    return traj + d_pos


def add_drift_to_traj(traj, drift_amplitude=0.0):
    drift, direction = get_drift(traj, drift_amplitude)
    return add_drift(traj, drift), direction * drift_amplitude


def add_noise(traj, noise_factor):
    noise_amplitude = noise_factor * get_traj_scale(traj)
    delta = np.random.normal(0, noise_amplitude, traj.shape)
    return traj + delta


def traj_scale(positions, scale_name):
    dr = get_steps(positions)
    if scale_name == "step_std":
        scale = np.std(dr)
    elif scale_name == "step_mean":
        scale = np.mean(dr)
    elif scale_name == "step_sum":
        scale = np.sum(dr)
    elif scale_name == "pos_std":
        # mean of variances of position along each dimension
        scale = np.sqrt(np.mean(np.var(positions, axis=0)))
    else:
        raise NotImplementedError(f"Unknown scale {scale_name}")
    if scale == 0:
        scale = 1.0
    return scale


def get_steps(positions):
    dr = positions[1:] - positions[:-1]
    return np.sqrt(np.sum(dr ** 2, axis=1))


class HiddenPrints:
    """
    https://stackoverflow.com/questions/8391411/how-to-block-calls-to-print
    """

    def __enter__(self):
        self._original_stdout = sys.stdout
        sys.stdout = open(os.devnull, "w")

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.close()
        sys.stdout = self._original_stdout



import numpy as np


def complete_graph(N):
    edge_start = np.array([i // N for i in range(N ** 2) if (i % N) != i // N])
    edge_end = np.array([i % N for i in range(N ** 2) if (i % N) != i // N])
    return edge_start, edge_end


def edges_uniform(N, D):
    delta = 1 + np.random.choice(N - 1, size=N * D, replace=True)
    edge_start = np.tile(np.arange(N), D)
    edge_end = (edge_start + delta) % N
    return edge_start, edge_end


def edges_geom_causal(N, D):
    # 0,0,0,...,N-1,N-1,N-1
    step_sizes = np.unique(np.floor(np.logspace(0, np.log10(N - 1), D)))
    # reduce mean degree to prevent doubled edges
    D = len(step_sizes)
    edge_start = np.repeat(np.arange(N), D)
    # A,B,C,...,A,B,C
    delta = -np.tile(step_sizes, N)
    edge_end = (edge_start + delta) % N

    return edge_start, edge_end


def traj_scale(positions: np.array, scale_name: str):
    dr, steps = get_steps(positions, return_steps=True)
    if scale_name == "step_std":
        scale = np.std(dr)
    elif scale_name == "step_mean":
        scale = np.mean(dr)
    elif scale_name == "msd":
        scale = np.sqrt(np.mean(dr ** 2) - np.sum(np.mean(steps) ** 2))
    elif scale_name == "step_sum":
        scale = np.sum(dr)
    elif scale_name == "pos_std":
        # mean of variances of position along each dimension
        scale = np.sqrt(np.mean(np.var(positions, axis=0)))
    else:
        raise NotImplementedError(f"Unknown scale {scale_name}")
    if scale == 0:
        scale = 1.0
    return scale


def get_steps(positions, return_steps=False):
    dr = positions[1:] - positions[:-1]
    if not return_steps:
        return np.sqrt(np.sum(dr ** 2, axis=1))
    else:
        return np.sqrt(np.sum(dr ** 2, axis=1)), dr



from torchmetrics.functional import confusion_matrix
from pytorch_lightning.callbacks import Callback
from .network_tools import is_concerned
from collections import defaultdict
from tqdm import tqdm
from ..data.datamodule import EMPTY_FIELD_VALUE
import os, sys
import torch
import numpy as np
import matplotlib.pyplot as plt


class Plotter(Callback):
    def on_fit_start(self, trainer, pl_module):
        self.max_capacity = int(1e4)
        self.tb = trainer.logger.experiment
        self.round = 0

    def on_validation_epoch_start(self, trainer, pl_module):
        self.preds = defaultdict(lambda: [])
        self.info = defaultdict(lambda: [])
        self.n_items = 0

    def on_validation_batch_end(
        self, trainer, pl_module, outputs, batch, batch_idx, dataloader_idx
    ):
        if self.n_items > self.max_capacity:
            return
        targets = outputs["targets"]
        preds = outputs["preds"]

        for i, param in enumerate(targets):
            self.n_items += targets[param].shape[0]
            self.info[param].append(targets[param])
            self.preds[param].append(preds[param])

    def on_validation_epoch_end(self, trainer, pl_module):

        for param in tqdm(self.info, leave=False, colour="blue"):
            # if param not in ["alpha", "drift_norm", "model"]:
            #    continue
            info = torch.cat(self.info[param], dim=0).cpu()
            pred = torch.cat(self.preds[param], dim=0).cpu()

            if param == "model":
                # info = info
                pred = torch.argmax(pred, dim=1)
                n_models = len(pl_module.hparams["RW_types"])
                CM = (
                    confusion_matrix(pred, info, n_models, normalize="true")
                    .detach()
                    .cpu()
                )

                fig = plt.figure()
                ax = fig.add_subplot(111)
                ax.imshow(CM, cmap="Blues", vmin=0.0, vmax=1.0)
                ax.set_xticks(np.arange(n_models))
                ax.set_xticklabels(pl_module.hparams["RW_types"])
                ax.set_yticks(np.arange(n_models))
                ax.set_yticklabels(pl_module.hparams["RW_types"])

                self.tb.add_figure("model_val", fig, global_step=self.round, close=True)

            elif param == "drift_norm":
                cond = is_concerned(info).detach().cpu().numpy()
                info = info[:, 0].detach().cpu().numpy()
                pred = pred[:, 0].detach().cpu().numpy()
                info[~cond] = 0.0

                plt.figure()
                plt.hist(pred[~cond])

                fig = plt.figure()
                ax = fig.add_subplot(211)
                ax.scatter(info[cond], pred[cond], label="BM & OU with drift", s=3)
                ax.plot([0.0, 0.5], [0.0, 0.5], ls=":", c="red")
                ax.set_xlim((-0.05, 0.55))
                ax.set_ylim((-0.05, 0.55))
                ax.set_xlabel("True drift")
                ax.set_ylabel("Inferred drift")
                ax.legend()

                ax = fig.add_subplot(212)
                ax.hist(
                    pred[~cond],
                    bins=30,
                    density=True,
                    label="Anomalous diffusion",
                    range=(-0.05, 1.0),
                    alpha=0.5,
                )
                ax.hist(
                    pred[cond & (info <= 0.25)],
                    bins=30,
                    density=True,
                    label="BM & OU w/ low drift",
                    range=(-0.05, 1.0),
                    alpha=0.5,
                )
                ax.hist(
                    pred[cond & (info > 0.25)],
                    bins=30,
                    density=True,
                    label="BM & OU w/ high drift",
                    range=(-0.05, 1.0),
                    alpha=0.5,
                )
                ax.set_xlabel("Inferred drift")
                ax.legend()

                plt.tight_layout()

                self.tb.add_figure("DNorm_val", fig, global_step=self.round, close=True)

            elif pred.shape[1] == info.shape[1]:
                fig = plt.figure()
                for d in range(pred.shape[1]):
                    plt.scatter(info[:, d], pred[:, d], label="dimension %d" % d, s=1)
                plt.xlabel("True")
                plt.ylabel("Predicted")
                if pred.shape[1] > 1:
                    plt.legend()
                plt.title(param)
                plt.tight_layout()
                self.tb.add_figure(
                    "%s_val" % param, fig, global_step=self.round, close=True
                )

        self.round += 1


class LatentSpaceSaver(Callback):
    def on_fit_start(self, trainer, pl_module):
        self.max_capacity = int(1e3)  # TB samples down to 5e3 points anyway
        self.tb = trainer.logger.experiment
        self.round = 0

    def on_validation_epoch_start(self, trainer, pl_module):
        self.latent_vectors = []
        self.info = defaultdict(lambda: [])
        self.n_items = 0

    def on_validation_batch_end(
        self, trainer, pl_module, outputs, batch, batch_idx, dataloader_idx
    ):
        if self.n_items > self.max_capacity:
            return
        h = outputs["preds"]["latent"]
        targets = outputs["targets"]

        self.n_items += h.shape[0]
        for param in targets:
            self.info[param].append(targets[param])
        self.latent_vectors.append(h)

    def on_validation_epoch_end(self, trainer, pl_module):
        self.round += 1

        # if not self.round % 10 == 0:
        #    return

        H = torch.cat(self.latent_vectors, dim=0).detach().cpu().numpy()
        for param in tqdm(self.info, leave=False, colour="blue"):
            if param not in ["alpha", "drift_norm", "model"]:
                continue
            info = torch.cat(self.info[param], dim=0)
            if param == "model":
                info = [pl_module.hparams["RW_types"][i] for i in info]
            else:
                info = [
                    "%.2f" % i if i != EMPTY_FIELD_VALUE else 0.0 for i in info[:, 0]
                ]
            try:
                self.tb.add_embedding(
                    H, metadata=info, tag=param, global_step=self.round
                )
            except AttributeError as e:
                print(e)
                print(
                    "Conflict between tensorboard and tensorflow (installed with umap-learn...)"
                )
                break

    def on_test_end(self, trainer, pl_module):
        metric_dict = dict()
        if "model" in pl_module.losses:
            metric_dict["model_F1"] = trainer.logged_metrics["test_F1"].item()
        if "alpha" in pl_module.losses:
            metric_dict["alpha_MAE"] = trainer.logged_metrics["test_MAE"].item()
        if "drift_norm" in pl_module.losses:
            metric_dict["drift_norm_EV"] = trainer.logged_metrics["test_EV"].item()
        hparams = dict()
        for p in pl_module.hparams:
            value = pl_module.hparams[p]
            if type(value) is tuple:
                if len(value) == 2:
                    hparams["%s_min" % p] = value[0]
                    hparams["%s_max" % p] = value[1]
            elif type(value) is list:
                for list_item in value:
                    hparams["has_%s" % list_item] = True
            else:
                hparams[p] = value
        self.tb.add_hparams(hparams, metric_dict, run_name="final")
        self.tb.flush()



import numpy as np
from ..simulation.traj_tools import *
from torch_geometric.data import Data
import torch
import numpy as np

from .data_tools import *

EMPTY_FIELD_VALUE = np.nan


class SkeletonTrajData(Data):
    def __init__(
        self,
        raw_positions,
        graph_info={},
        traj_info={},
        original_positions=None,
        **kwargs,
    ):

        # if raw_positions is None:
        #    print("raw_positions was none")
        #    raw_positions = np.random.normal(size=(10, 2))
        try:
            dim = raw_positions.shape[1]

            if original_positions is None:
                original_positions = raw_positions
            if not type(original_positions) is np.array:
                original_positions = np.array(original_positions).copy()

            default_traj_info = {
                "model": "unknown",
                "model_index": 0,
                "log_theta": EMPTY_FIELD_VALUE,
                "alpha": EMPTY_FIELD_VALUE,
                "noise": 0.0,
                "tau": np.inf,
                "seed": EMPTY_FIELD_VALUE,
                "log_diffusion": EMPTY_FIELD_VALUE,
            }

            for key in default_traj_info:
                if default_traj_info[key] is None:
                    continue
                if key not in traj_info:
                    traj_info[key] = default_traj_info[key]

            raw_positions -= raw_positions[0]
            positions, clipped_steps = self.safe_positions(raw_positions, graph_info)

            edge_index = self.get_edges(positions.shape[0], graph_info)

            reshape = lambda t: torch.reshape(torch.from_numpy(t), (1, -1))
            float_to_torch = lambda t: torch.Tensor([t]).view((1, 1))

            # print("Skeleton seed = %d" % traj_info["seed"])

            super(SkeletonTrajData, self).__init__(
                pos=torch.from_numpy(positions).float(),
                original_pos=torch.from_numpy(original_positions).float(),
                clipped_steps=float_to_torch(clipped_steps),
                edge_index=edge_index,
                length=float_to_torch(positions.shape[0]),
                alpha=float_to_torch(traj_info["alpha"]),
                log_theta=float_to_torch(traj_info["log_theta"]),
                log_tau=reshape(np.asarray(np.log10(traj_info["tau"]))),
                model=float_to_torch(traj_info["model_index"]).long(),
                noise=float_to_torch(traj_info["pos_uncertainty"]),
                seed=torch.Tensor([int(traj_info["seed"])]),
                log_diffusion=float_to_torch(traj_info["log_diffusion"]),
            )
        except AttributeError as e:
            # print(e)
            super(SkeletonTrajData, self).__init__(
                pos=torch.from_numpy(np.random.normal(size=(10, 2)))
            )
        self.coalesce()

    def safe_positions(self, positions, graph_info):
        # clips too long jumps, returns the number of clipped steps
        if graph_info["clip_trajs"] == True:
            dr = get_steps(positions)
            M = np.median(dr) + 10 * np.std(dr)
            clipped_steps = np.sum(dr > M)
            dr_clipped = np.clip(dr, a_min=0.0, a_max=M)
            clipped_positions = np.zeros(positions.shape)
            deltas = positions[1:] - positions[:-1]
            dr[dr == 0.0] = 1.0  # in case
            ratio = dr_clipped / dr
            ratio[np.isnan(ratio)] = 0.0
            deltas = deltas * np.reshape(ratio, (-1, 1))
            clipped_positions[1:] = np.cumsum(deltas, axis=0)
            return clipped_positions, clipped_steps
        else:
            return positions, 0

    @classmethod
    def get_edges(cls, N, graph_info):

        D = graph_info["edges_per_point"]
        if D >= N:
            edge_start, edge_end = complete_graph(N)
        if graph_info["edge_method"] == "uniform":
            edge_start, edge_end = edges_uniform(N, D)
        elif graph_info["edge_method"] == "geom_causal":
            edge_start, edge_end = edges_geom_causal(N, D)
        else:
            raise NotImplementedError(f"Method { graph_info['edge_method'] } not known")
        e = np.stack([edge_start, edge_end], axis=0)
        return torch.from_numpy(e).long()



import pytorch_lightning as pl
from codecarbon import track_emissions
from gratin.models.network_lightning import MainNet
from gratin.data.data_classes import DataModule
from pytorch_lightning.callbacks import (
    EarlyStopping,
    LearningRateMonitor,
    ModelCheckpoint,
)
from gratin.training.callbacks import LatentSpaceSaver, Plotter


def setup(tasks, ds_params, net_params, dl_params, graph_info):
    pl.seed_everything(1234)
    dm = DataModule(ds_params=ds_params, dl_params=dl_params, graph_info=graph_info)
    dm.setup()
    model = MainNet(
        tasks=tasks, latent_dim=net_params["latent_dim"], n_c=net_params["n_c"], dm=dm
    )

    return model, dm


@track_emissions(project_name="unsupervised-rw")
def train_and_test(model, dm, logger):

    ES = EarlyStopping(
        monitor="train_loss",
        min_delta=0.001,
        patience=10,
        verbose=True,
        mode="min",
        strict=True,
    )
    LRM = LearningRateMonitor("epoch")
    tag = "unsupervised_v1"
    CKPT = ModelCheckpoint(
        dirpath="/gaia/models",
        filename=tag,
        monitor="train_loss",
        mode="min",
        verbose=True,
    )
    LSS = LatentSpaceSaver()

    trainer = pl.Trainer(
        auto_lr_find=False,
        gpus=1,
        auto_select_gpus=False,
        gradient_clip_val=1.0,
        log_gpu_memory="min_max",
        reload_dataloaders_every_epoch=True,
        callbacks=[ES, LRM, CKPT, LSS, Plotter()],
        accelerator="ddp",
        log_every_n_steps=150,
        flush_logs_every_n_steps=300,
        terminate_on_nan=True,
        track_grad_norm=2,
        weights_summary="full",
        profiler="simple",
        replace_sampler_ddp=True,
        logger=logger,
    )
    # trainer.tune(model=model, datamodule=dm)
    trainer.fit(model=model, datamodule=dm)
    trainer.test(model=model, datamodule=dm, verbose=True)


if __name__ == "__main__":
    ds_params = {
        "dim": 2,
        "RW_types": ["sBM", "CTRW", "fBM", "LW", "OU", "BM"],
        "N": int(1e5),
        "length_range": (10, 30),
        "noise_range": (0.2, 0.5),  # Issues when noise is too low
        "drift_range": (0.0, 0.5),
    }

    net_params = {"latent_dim": 32, "n_c": 32}

    dl_params = {"batch_size": 128, "num_workers": 10}

    graph_info = {
        "edges_per_point": 20,
        "clip_trajs": True,
        "scale_types": ["step_std", "step_mean", "pos_std"],
        "log_features": True,
        "edge_method": "geom_causal",
    }

    tasks = ["model", "drift_norm", "alpha"]

    model, dm = setup(tasks, ds_params, net_params, dl_params, graph_info)

    tb_logger = pl.loggers.TensorBoardLogger(
        save_dir="/gaia/tb_logs",
        default_hp_metric=False,
    )
    train_and_test(model, dm, tb_logger)


# Copyright 2019 by Gorka Munoz-Gil under the MIT license.
# This file is part of the Anomalous diffusion challenge (AnDi), and is
# released under the "MIT License Agreement". Please see the LICENSE
# file that should have been included in the repository containing the file
# (github.com/gorkamunoz/ANDI)

import numpy as np

__all__ = ["bm1D", "regularize", "sample_sphere", "normalize"]


def bm1D(T, D, deltaT=False):
    """Creates a 1D Brownian motion trajectory"""
    if D < 0:
        raise ValueError("Only positive diffusion coefficients allowed.")
    if not deltaT:
        deltaT = 1
    return np.cumsum(np.sqrt(2 * D * deltaT) * np.random.randn(int(T)))


def regularize(positions, times, T):
    """Regularizes a trajectory with irregular sampling times.
    Arguments:
        - positions (numpy.array): collections of positions
        - times (numpy.array): collections of sampling times
        - T (int): lenght of the output trajectory

        return: numpy.array"""
    times = np.append(0, times)
    pos_r = np.zeros(T)
    for idx in range(len(times) - 1):
        pos_r[int(times[idx]) : int(times[idx + 1])] = positions[idx]
    pos_r -= pos_r[0]
    return pos_r


def sample_sphere(N, R):
    """Samples random number that lay in the surface of a 3D sphere centered in
    zero and with radius R
    Arguments:
        - N (int): number of points to generate.
        - R (int or numpy.array): radius of the sphere. If int, all points have
        the same radius, if numpy.array, each number has different radius."""
    vecs = np.random.randn(3, N)
    vecs /= np.linalg.norm(vecs, axis=0)
    return R * vecs


def normalize(trajs):
    """Normalizes trajectories by substracting average and dividing by sqrt of msd
    Arguments:
        - traj: trajectory or ensemble of trajectories to normalize.
    - dimensions: dimension of the trajectory.
        return: normalized trajectory"""
    if len(trajs.shape) == 1:
        trajs = np.reshape(trajs, (1, trajs.shape[0]))
    trajs = trajs - trajs.mean(axis=1, keepdims=True)
    displacements = (trajs[:, 1:] - trajs[:, :-1]).copy()
    variance = np.std(displacements, axis=1)
    variance[variance == 0] = 1
    new_trajs = np.cumsum((displacements.transpose() / variance).transpose(), axis=1)
    return np.concatenate((np.zeros((new_trajs.shape[0], 1)), new_trajs), axis=1)



import torch
import torch.nn as nn
import numpy as np
from .diverse import MLP


class ACB(nn.Module):
    def __init__(self, dim_theta, dim_mu, stable_s=False, alpha=1.7, s_abs=4, t_abs=4):
        """
        dim_theta : dimension of the parameters vector
        dim_mu : dimension of the summary vector
        """
        super(ACB, self).__init__()

        self.n_1 = int(dim_theta // 2)
        self.n_2 = int(np.ceil(dim_theta / 2))
        assert self.n_1 + self.n_2 == dim_theta

        self.s_limit = 0.0

        weight_norm = True  # "Density estimation using real NVP" paper uses this
        residual = False
        s_params = {
            "residual": residual,
            "use_weight_norm": weight_norm,
            "out_range": (-s_abs, s_abs),
            "activation": "ELU",
            "use_batch_norm": False,
        }
        t_params = {
            "residual": residual,
            "use_weight_norm": weight_norm,
            "out_range": (-t_abs, t_abs),
            "activation": "ELU",
            "use_batch_norm": False,
        }

        self.permutation = nn.Parameter(torch.randperm(dim_theta), requires_grad=False)
        self.inv_permutation = nn.Parameter(
            torch.argsort(self.permutation), requires_grad=False
        )

        self.s_1 = MLP(
            [
                self.n_2 + dim_mu,
                64,
                64,
                64,
                self.n_1,
            ],
            **s_params
        )
        self.t_1 = MLP(
            [
                self.n_2 + dim_mu,
                64,
                64,
                64,
                self.n_1,
            ],
            **t_params
        )
        self.s_2 = MLP(
            [
                self.n_1 + dim_mu,
                64,
                64,
                64,
                self.n_2,
            ],
            **s_params
        )
        self.t_2 = MLP(
            [
                self.n_1 + dim_mu,
                64,
                64,
                64,
                self.n_2,
            ],
            **t_params
        )
        self.stable_s = stable_s

        self.alpha = alpha

    def clamp(self, s):
        return (2.0 * self.alpha / np.pi) * torch.atan(s / self.alpha)

    def clamp_loss(self, s):
        #
        #mean_before = torch.mean(s)
        s = torch.clamp(s, max=self.s_limit, min=-np.inf)
        #mean_after = torch.mean(s)
        #print(
        #    "s_lim = %.3f\t Mean before %.2f \t after %.2f"
        #    % (self.s_limit, mean_before.item(), mean_after.item())
        #)
        return s

    def forward(self, theta, x, inverse=False, log_det_J=False):

        if not inverse:
            u = torch.index_select(theta, 1, self.permutation)

            u_1 = torch.index_select(u, 1, torch.arange(self.n_1, device=u.device))
            u_2 = torch.index_select(
                u, 1, torch.arange(self.n_2, device=u.device) + self.n_1
            )

            U_2 = torch.cat((u_2, x), dim=1)
            s_1 = self.clamp(self.s_1(U_2))
            if self.stable_s:
                s_1 = s_1 - torch.mean(s_1, dim=1, keepdim=True)
            v_1 = (u_1 * torch.exp(s_1)) + self.t_1(U_2)

            V_1 = torch.cat((v_1, x), dim=1)
            s_2 = self.clamp(self.s_2(V_1))
            if self.stable_s:
                s_2 = s_2 - torch.mean(s_2, dim=1, keepdim=True)
            v_2 = (u_2 * torch.exp(s_2)) + self.t_2(V_1)

            result = torch.cat((v_1, v_2), dim=1)
            log_det_J_loss = torch.sum(self.clamp_loss(s_1), dim=1) + torch.sum(
                self.clamp_loss(s_2), dim=1
            )

            if log_det_J:
                return result, log_det_J_loss
            else:
                return result

        else:
            v = theta
            v_1 = torch.index_select(v, 1, torch.arange(self.n_1, device=v.device))
            v_2 = torch.index_select(
                v, 1, torch.arange(self.n_2, device=v.device) + self.n_1
            )

            V_1 = torch.cat((v_1, x), dim=1)
            u_2 = (v_2 - self.t_2(V_1)) * torch.exp(-self.clamp(self.s_2(V_1)))

            U_2 = torch.cat((u_2, x), dim=1)
            u_1 = (v_1 - self.t_1(U_2)) * torch.exp(-self.clamp(self.s_1(U_2)))

            inv = torch.cat((u_1, u_2), dim=1)

            result = torch.index_select(inv, 1, self.inv_permutation)

            return result


class InvertibleNet(nn.Module):
    def __init__(
        self,
        dim_theta,
        dim_x,
        n_blocks,
        stable_s=False,
        alpha_clip=1.7,
        t_abs=4,
        s_abs=4,
    ):
        super(InvertibleNet, self).__init__()
        self.ACBs = nn.Sequential(
            *[
                ACB(
                    dim_theta,
                    dim_x,
                    stable_s=stable_s,
                    alpha=alpha_clip,
                    t_abs=t_abs,
                    s_abs=s_abs,
                )
                for k in range(n_blocks)
            ]
        )

    def set_s_lim(self, s_lim):
        for acb in self.ACBs:
            acb.s_limit = s_lim

    def forward(self, theta, x, inverse=False, return_intermediates=False):
        if not inverse:
            log_J_total = 0.0
            z_intermediates = [theta]
            for acb in self.ACBs:
                theta, log_J = acb(theta, x, inverse=False, log_det_J=True)
                z_intermediates.append(theta.clone())
                log_J_total = log_J_total + log_J
            z = theta
            if not return_intermediates:
                return z, log_J_total
            else:
                return z, log_J_total, z_intermediates
        else:
            theta_intermediates = [theta]
            z = theta
            for acb in self.ACBs[::-1]:
                z = acb(z, x, inverse=True, log_det_J=False)
                theta_intermediates.append(z.clone())
            theta = z
            if not return_intermediates:
                return theta
            else:
                return theta, theta_intermediates



from torch_geometric.loader import DataLoader
import pytorch_lightning as pl
from torch.utils.data import random_split
from torch.utils.data.distributed import DistributedSampler
from . import default_ds_params, default_graph_info

# from ..simulation.diffusion_models import generators, params_sampler
from ..simulation.traj_tools import *

from .data_tools import *
from .dataset import *

EMPTY_FIELD_VALUE = -999


class DataModule(pl.LightningDataModule):
    def __init__(self, dl_params={}, ds_params={}, graph_info={}):

        super().__init__()

        self.ds_params = default_ds_params
        self.graph_info = default_graph_info
        self.ds_params.update(ds_params)
        self.graph_info.update(graph_info)
        self.dl_params = dl_params

        self.batch_size = dl_params["batch_size"]
        self.epoch_count = 1

        self.round = 0
        self.exclude_keys = ["raw_positions"]

    def setup(self, stage=None, plot=True):
        if stage is None:
            print("stage is None, strange...")
        self.recreate_datasets(stage=stage, plot=plot)

    def recreate_datasets(self, stage="fit", plot=True):
        ds_params = {
            "N": self.ds_params["N"],
            "dim": self.ds_params["dim"],
            "noise_range": self.ds_params["noise_range"],
            "model_types": self.ds_params["RW_types"],
            "logdiffusion_range": self.ds_params["logdiffusion_range"],
            "length_range": self.ds_params["length_range"],
            "time_delta": self.ds_params["time_delta"],
        }  # a bit redundant, but we recreate a ds_params, just to make sure it has only good arguments

        if stage == "fit" or stage is None:
            self.ds_train = TrajDataSet(
                **ds_params,
                graph_info=self.graph_info,
                # seed_offset=0
                seed_offset=self.ds_params["N"] * self.epoch_count,
            )
        if self.round == 0:
            ds_params_val = dict(ds_params)
            ds_params_val["N"] = 3000
            self.ds_val = TrajDataSet(
                **ds_params_val,
                graph_info=self.graph_info,
                seed_offset=0,
            )

        if stage == "test" or stage is None:
            ds_params_test = dict(ds_params)
            ds_params_test["N"] //= 10
            ds = TrajDataSet(
                **ds_params_test,
                graph_info=self.graph_info,
                # seed_offset=0
                seed_offset=self.ds_params["N"] * (self.epoch_count + 1),
            )
            self.ds_test = ds

        if self.round == 0 and plot:
            print("Plotting")
            self.traj_examples = ds.make_plot()
        self.round += 1

    def train_dataloader(self):
        # print(f"Call train_dataloader for the {self.epoch_count}th time")
        self.epoch_count += 1
        self.recreate_datasets(stage="fit")

        return DataLoader(
            self.ds_train,
            num_workers=self.dl_params["num_workers"],
            drop_last=True,
            batch_size=self.batch_size,
            exclude_keys=self.exclude_keys,
            # sampler=DistributedSampler(self.ds_train) if not self.no_parallel else None,
            pin_memory=True,
        )

    def val_dataloader(self):
        return DataLoader(
            self.ds_val,
            num_workers=self.dl_params["num_workers"],
            drop_last=True,
            batch_size=self.batch_size,
            exclude_keys=self.exclude_keys,
            # sampler=DistributedSampler(self.ds_val) if not self.no_parallel else None,
            pin_memory=True,
        )

    def test_dataloader(self):
        self.recreate_datasets(stage="test")

        return DataLoader(
            self.ds_test,
            num_workers=self.dl_params["num_workers"],
            batch_size=self.batch_size,
            drop_last=True,
            exclude_keys=self.exclude_keys,
            # sampler=DistributedSampler(self.ds_test),
            pin_memory=True,
        )



import torch_geometric.transforms as Transforms
from torch_geometric.data import Dataset
import numpy as np
import warnings
from ..simulation.diffusion_models import generators, params_sampler
from ..simulation.traj_tools import *
from ..data.data import TrajData
import matplotlib.pyplot as plt
import numpy as np
from scipy.stats import truncnorm

EMPTY_FIELD_VALUE = -999


class TrajDataSet(Dataset):
    def __init__(
        self,
        N: int,
        dim: int,
        graph_info: dict,
        length_range: tuple,  # e.g = (5,50)
        noise_range: tuple,  # e.g. = (0.1,0.5),
        model_types: list,  # e.g. = ["fBM","CTRW"],
        seed_offset: int,  # e.g. = 0):
        time_delta: float,
        logdiffusion_range: tuple = (-2.0, 0.5),
    ):
        self.N = N

        self.seed_offset = seed_offset
        print("Create TrajDataset, seed_offset = %d" % seed_offset)
        self.graph_info = graph_info

        self.generators = generators[dim]
        self.model_types = model_types
        self.length_range = length_range
        self.noise_range = noise_range
        self.dim = dim
        self.log_diff_range = logdiffusion_range  # (-2.0, 0.5)
        self.time_delta = time_delta
        super(TrajDataSet, self).__init__(transform=Transforms.ToSparseTensor())

    def len(self):
        return self.N

    def generate_log_diffusion(self):
        OK = False
        # log_diffusion = np.random.uniform(*self.log_diff_range)
        diff_mean = 0.5 * (self.log_diff_range[0] + self.log_diff_range[1])
        diff_std = 0.3 * (self.log_diff_range[1] - self.log_diff_range[0])
        a, b = (self.log_diff_range[0] - diff_mean) / diff_std, (
            self.log_diff_range[1] - diff_mean
        ) / diff_std
        while not OK:
            log_diffusion = truncnorm.rvs(
                a,
                b,
                loc=diff_mean,
                scale=diff_std,
            )
            OK = (log_diffusion >= self.log_diff_range[0]) and (
                log_diffusion <= self.log_diff_range[1]
            )
        return log_diffusion

    def get_traj_params(self):
        model_index = np.random.choice(len(self.model_types))
        model = self.model_types[model_index]
        params = params_sampler(model, seed=None)
        # length = np.random.randint(low=self.length_range[0], high=self.length_range[1])
        length = int(
            np.power(
                10,
                np.random.uniform(
                    low=np.log10(self.length_range[0]),
                    high=np.log10(self.length_range[1]),
                ),
            )
        )

        log_diffusion = self.generate_log_diffusion()

        diffusion = np.power(10.0, log_diffusion)

        pos_uncertainty = np.random.uniform(self.noise_range[0], self.noise_range[1])
        # noise_factor = np.random.uniform(self.noise_range[0], self.noise_range[1])

        trajs_params = {
            "model_index": model_index,
            "model": model,
            "params": params,
            "length": length,
            "diffusion": diffusion,
            "log_diffusion": log_diffusion,
            "pos_uncertainty": pos_uncertainty,
        }

        return trajs_params

    def get_raw_traj(self, traj_params):

        OK = False
        counter = 0

        while not OK:
            raw_pos = self.generators[traj_params["model"]](
                T=traj_params["length"], **traj_params["params"]
            )
            raw_pos -= raw_pos[0]
            OK = np.isnan(raw_pos).sum() == 0
            # OK = OK and (np.max(raw_pos[:, 0]) > np.min(raw_pos[:, 0]))
            OK = OK or (traj_params["model"] == "empty")
            if not OK:
                counter += 1
                # print("raw_pos has nans %d" % counter)
                # print(params)
        if counter > 10:
            # print(
            #    "More than 10 trials (%d) to generate a %s"
            #    % (counter, traj_params["model"])
            # )
            # for p in traj_params["params"]:
            #    print(p, traj_params["params"][p])
            pass

        assert raw_pos is not None, "raw_pos is None %s" % raw_pos

        return raw_pos

    def apply_diffusion_and_noise(self, raw_pos, traj_params):

        diffusion = traj_params["diffusion"]
        # force_norm = EMPTY_FIELD_VALUE

        raw_pos *= np.sqrt(diffusion * self.time_delta)

        if traj_params["pos_uncertainty"] > 0.0:
            noisy_pos = raw_pos + traj_params["pos_uncertainty"] * np.random.randn(
                *raw_pos.shape
            )
        else:
            noisy_pos = raw_pos
        return noisy_pos

    def get_traj(self, seed, return_raw=False):

        # print("Get traj %d" % seed)

        np.random.seed(seed)
        traj_params = self.get_traj_params()
        traj_params["seed"] = seed
        # print(traj_params)
        # print(traj_params["log_diffusion"])
        raw_pos = self.get_raw_traj(traj_params)
        pos = self.apply_diffusion_and_noise(raw_pos, traj_params)

        assert pos is not None, "pos is None %s" % traj_params

        if not return_raw:
            return pos, traj_params
        else:
            return pos, traj_params, raw_pos

    def make_plot(self):
        fig = plt.figure(figsize=(10, 10), dpi=150)

        traj_params = {"length": 20, "diffusion": 1e-3, "pos_uncertainty": 0.1}

        for i in range(10):

            for j in range(10):
                if i <= 3:
                    traj_params["model"] = "BM"
                    traj_params["diffusion"] = np.power(10.0, self.log_diff_range[1])
                    traj_params["params"] = {"alpha": 1.0}
                    traj_params["pos_uncertainty"] = 0.0
                else:
                    traj_params = self.get_traj_params()

                model = traj_params["model"]
                raw_pos = self.get_raw_traj(traj_params)
                noisy_pos = self.apply_diffusion_and_noise(raw_pos, traj_params)

                ax = fig.add_subplot(10, 10, j * 10 + i + 1)
                if self.dim == 1:
                    ax.plot(noisy_pos[:, 0])
                else:
                    ax.plot(noisy_pos[:, 0], noisy_pos[:, 1])
                params = traj_params["params"]
                pos_uncertainty = traj_params["pos_uncertainty"]
                if "tau" in params:
                    desc = "$\\tau_c$ = %d" % params["tau"]
                elif "alpha" in params:
                    desc = f"{model[:4]} - $\\alpha$:{float(params['alpha']):{1}.{2}} - N{pos_uncertainty:{1}.{1}}"
                else:
                    desc = f"{model[:4]} - N{pos_uncertainty:{1}.{1}}"
                ax.set_title(desc, fontsize=6)
                if self.dim >= 2:
                    ax.set_aspect("equal")
                plt.axis("off")
        plt.tight_layout()
        return fig

    def get(self, idx):
        # pos = np.zeros((10,2))

        # if self.train:
        #    print("Train index %d" % idx)
        # else:
        #    print("Test index %d" % idx)
        # print("Get idx = %d" % idx)
        seed = idx + self.seed_offset
        # print("generating seed = %d " % seed)
        noisy_pos, traj_params, raw_pos = self.get_traj(seed=seed, return_raw=True)
        assert noisy_pos is not None, "noisy_pos is None : %s" % traj_params
        traj_info = traj_params["params"]
        traj_params.pop("params")
        traj_info.update(traj_params)
        # print("traj_info : %d" % traj_info["seed"])
        return TrajData(
            noisy_pos,
            graph_info=self.graph_info,
            traj_info=traj_info,
            original_positions=raw_pos,
        )


class ExpTrajDataSet(Dataset):
    def __init__(
        self,
        dim: int,
        graph_info: dict,
        trajs: list,
    ):  # e.g. = 0):
        self.N = len(trajs)
        self.trajs = trajs
        self.graph_info = graph_info
        self.dim = dim

        transform = Transforms.ToSparseTensor()

        super(ExpTrajDataSet, self).__init__(transform=transform)

    def len(self):
        return self.N

    def make_plot(self, info=None):
        fig = plt.figure(figsize=(10, 10), dpi=150)

        N = 100
        if info is not None:
            N = len(info)
        for i in range(min(N, len(self))):
            noisy_pos = self.trajs[i]
            ax = fig.add_subplot(10, 10, i + 1)
            if self.dim == 1:
                ax.plot(noisy_pos[:, 0])
            else:
                ax.plot(noisy_pos[:, 0], noisy_pos[:, 1])
            ax.set_aspect("equal")
            if info is not None:
                ax.set_title("info : %.2f" % info[i])
            plt.axis("off")
        plt.tight_layout()
        return fig

    def get(self, idx):

        noisy_pos = self.trajs[idx]
        traj_info = {"index": idx, "pos_uncertainty": 0.03, "seed": idx}
        return TrajData(
            noisy_pos,
            graph_info=self.graph_info,
            traj_info=traj_info,
            original_positions=None,
        )



default_graph_info = {
    "edges_per_point": 10,
    "scale_types": ["step_std"],
    "edge_method": "geom_causal",
    "data_type": "no_features",
}

default_ds_params = {
    "dim":2,
    "noise_range": (0.02, 0.05),
    "RW_types": ["fBM", "LW", "sBM", "OU", "CTRW"],
    "length_range": (7, 35),
    "logdiffusion_range":(-2.5,1.1),
    "N": int(1e5),
    "time_delta": 0.03,
}



import torch
import pytorch_lightning as pl
from pytorch_lightning.callbacks import (
    EarlyStopping,
    LearningRateMonitor,
    ModelCheckpoint,
)
from .callbacks import LatentSpaceSaver, Plotter
from ..models.main_net import MainNet
from ..data.datamodule import DataModule


def setup_model_and_dm(tasks, ds_params, net_params, dl_params, graph_info):
    pl.seed_everything(1234)
    dm = DataModule(ds_params=ds_params, dl_params=dl_params, graph_info=graph_info)
    dm.setup()
    model = MainNet(
        tasks=tasks, latent_dim=net_params["latent_dim"], n_c=net_params["n_c"], dm=dm
    )

    return model, dm


def setup_trainer(logger, dirpath="/gaia/models", tag="default"):
    ES = EarlyStopping(
        monitor="train_loss",  # We do not care about overfitting, train loss is more stable than validation
        min_delta=0.001,
        patience=5,
        verbose=True,
        mode="min",
        strict=True,
    )
    LRM = LearningRateMonitor("epoch")
    CKPT = ModelCheckpoint(
        dirpath=dirpath, filename=tag, monitor="train_loss", verbose=True, mode="min"
    )
    PLT = Plotter()

    trainer = pl.Trainer(
        auto_select_gpus=True,
        gradient_clip_val=1.0,
        reload_dataloaders_every_n_epochs=1,
        callbacks=[ES, LRM, CKPT, PLT],
        log_every_n_steps=150,
        max_epochs=100,
        detect_anomaly=True,
        track_grad_norm=2,
        logger=logger,
    )
    return trainer



import torch
import torch.nn as nn
import numpy as np


class fBMGenerator(nn.Module):
    def __init__(self, dim, **kwargs):
        self.dim = dim
        super(fBMGenerator, self).__init__(**kwargs)

    def f_(self, DT, alpha):
        return 0.5 * (
            torch.pow(torch.abs(DT + 1), alpha)
            + torch.pow(torch.abs(DT - 1), alpha)
            - 2 * torch.pow(torch.abs(DT), alpha)
        )

    def f(self, DT, alpha, tau, lam=1):
        R = self.f_(DT, alpha)
        R = torch.where(
            torch.abs(DT) >= tau, R * torch.exp(-lam * (torch.abs(DT) - tau)), R
        )
        return R

    def get_dx_cov(self, alpha, T, tau=np.inf, lam=1.0):

        BS = alpha.shape[0]
        t = torch.arange(T - 1, device=alpha.device, requires_grad=False).view(1, -1)
        # t : (1,T-1)
        t = t.expand(BS, -1)
        # t : (BS,T-1)
        t = t.view(BS, T - 1, 1)
        # t: (BS,T-1,1)
        i = t
        j = torch.transpose(t, 1, 2)
        k = j - i

        C = self.f(k, alpha.view(-1, 1, 1).double(), tau.view(-1, 1, 1), lam=lam)
        # min_EV = np.min(np.linalg.eigvals(C))
        # assert min_EV > 0.0, "min EV %.2f, alpha = %.2f, tau = %d" % (min_EV, alpha, tau)

        return C.double()

    def forward(self, alpha, tau, diffusion, T):
        # alpha : array de taille (BS)
        # tau : array de taille (BS)
        # T : int

        BS = alpha.shape[0]
        C = self.get_dx_cov(alpha, T=T, tau=tau, lam=1.0)
        try:
            # Returns a lower triangular matrix
            L = torch.cholesky(C).float()
        except Exception as e:
            print("alpha =")
            print(alpha)
            print("tau = ")
            print(tau)
            print("T = ")
            print(T)
            print(e)
            raise

        du = torch.randn(
            (BS, T - 1, self.dim), device=alpha.device, requires_grad=False
        )
        du = du * (torch.sqrt(diffusion).view(BS, 1, 1).expand(du.size()))

        dx = L @ du
        return torch.cat(
            [
                torch.zeros((BS, 1, self.dim), device=alpha.device),
                torch.torch.cumsum(dx, dim=1),
            ],
            dim=1,
        )



from collections import defaultdict
import pytorch_lightning as pl
from pytorch_lightning.metrics.regression import MeanAbsoluteError as MAE
from pytorch_lightning.metrics.regression import MeanSquaredError as MSE
from pytorch_lightning.metrics import F1
from pytorch_lightning.metrics import ExplainedVariance as EV
import torch.nn as nn
import torch
from functools import partial
from ..layers.diverse import MLP, AlphaPredictor, batch_from_positions
from ..layers.features_init import *
from ..training.network_tools import L2_loss, Category_loss, is_concerned
from ..data.data_classes import DataModule

from ..layers.encoders import *
from ..layers.InvNet import InvertibleNet

from torch.optim.lr_scheduler import ExponentialLR
from torch_geometric.data import Batch


class SelfSup(pl.LightningModule):
    def __init__(
        self,
        n_c: int,  # number of convolution channels
        latent_dim: int,
        dim: int,  # traj dim
        gamma: float = 0.98,
        lr: float = 1e-3,
        scale_types: list = ["step_std"],
        degree: int = 20,
        L_pred: int = 10,
    ):
        super().__init__()

        self.save_hyperparameters()
        x_dim = TrajsFeatures.x_dim(scale_types)
        e_dim = TrajsFeatures.e_dim(scale_types)
        self.save_hyperparameters({"x_dim": x_dim, "e_dim": e_dim})

        self.features_maker = TrajsFeatures()

        self.summary_net = TrajsEncoder2(
            n_c=n_c,
            latent_dim=latent_dim,
            x_dim=self.hparams["x_dim"],
            e_dim=self.hparams["e_dim"],
            traj_dim=0,  # On se fiche de l'orientation
            n_scales=0,  # On se fiche de la diffusivité
        )

        self.steps_pred = InvertibleNet(
            (L_pred - 1) * dim, latent_dim, n_blocks=3, stable_s=True
        )

    def forward(self, x):

        X, E, scales, orientation = self.features_maker(
            x, scale_types=self.hparams["scale_types"]
        )
        x.adj_t = x.adj_t.set_value(E)
        x.x = X
        # x.scales = scales
        # x.orientation = orientation

        assert x.x.shape[1] == self.hparams["x_dim"]
        h_true = self.summary_net(x)

        L = self.hparams["L_pred"]
        D = self.hparams["dim"]
        N = h_true.shape[0]

        # z = torch.randn((N, (L - 1) * D), device=self.device)
        # raw_steps = self.steps_pred(torch.cat((h_true), dim=1))
        epsilon = torch.randn((N, (L - 1) * D), device=h_true.device)
        raw_steps, log_J = self.steps_pred(epsilon, h_true)
        steps = raw_steps.view((N, D, L - 1))

        # assert torch.equal(raw_steps[0, (L - 1) * D - 1], steps[0, D - 1, L - 2])
        # assert torch.equal(raw_steps[1, (L - 1) * D - 1], steps[1, D - 1, L - 2])
        # if D >= 2:
        #    assert torch.equal(
        #        raw_steps[1, (L - 1) * (D - 1) - 1], steps[1, D - 2, L - 2]
        #    )
        pos = torch.cat(
            (
                torch.zeros((N, D, 1), device=raw_steps.device),
                torch.cumsum(steps, dim=2),
            ),
            dim=2,
        )
        pos = torch.transpose(pos, 1, 2)

        x_pred = batch_from_positions(pos, N=N, L=L, D=D, degree=self.hparams["degree"])

        (X_pred, E_pred, scales, orientation) = self.features_maker(
            x_pred,
            scale_types=self.hparams["scale_types"],
            return_intermediate=False,
        )
        # Need to re-create a batch to set the edge features values...
        x_pred = Batch(
            x_pred.batch, adj_t=x_pred.adj_t.set_value(E_pred), pos=x_pos, x=X_pred
        )

        h_pred = self.summary_net(x_pred)

        return h_true, h_pred, x_pos, log_J

    def training_step(self, batch, batch_idx):
        h_true, h_pred, traj_pred, log_J = self(batch)
        sigma_2 = 0.5 * torch.min(torch.var(h_true, dim=0) + torch.var(h_pred, dim=0))
        log_distance = torch.log(
            1 - torch.mean(nn.functional.cosine_similarity(h_true, h_pred))
        )
        # log_distance = torch.mean(torch.log(1e-5 + (h_true - h_pred) ** 2))
        h_norm = torch.mean(h_true ** 2 + h_pred ** 2)
        l = (
            log_distance
            + h_norm
            + (1.0 / (1e-5 + sigma_2))
            - 1.0 / torch.sqrt(sigma_2 + 1e-5)
            # + torch.mean(log_J ** 2)
        )

        debug = False
        if debug:
            for n, p in self.named_parameters():
                if p.grad is None:
                    print(n, p.grad)
                elif torch.sum(torch.isnan(p.grad)) == 0:
                    print(n, "numeric")
                else:
                    print(n, "============  NULL")

        self.log("training_loss", value=l, on_epoch=True, prog_bar=True)
        self.log("log_distance", value=log_distance, prog_bar=True)
        self.log("inv_sigma_2", value=1.0 / sigma_2, prog_bar=True)
        self.log("h_norm", value=h_norm, prog_bar=True)
        # self.log("log_J", value=torch.mean(log_J), prog_bar=True) # 0 if stable_s

        return {"loss": l}

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(
            self.parameters(), amsgrad=True, lr=self.hparams["lr"]
        )
        scheduler = ExponentialLR(optimizer, gamma=self.hparams["gamma"])
        return [optimizer], [scheduler]

    def optimizer_step(
        self,
        epoch,
        batch_idx,
        optimizer,
        optimizer_idx,
        optimizer_closure,
        on_tpu,
        using_native_amp,
        using_lbfgs,
    ):
        for n, p in self.named_parameters():
            if p.grad is None:
                continue
            if torch.sum(torch.isnan(p.grad)) > 0:
                print("%s has a nan gradient" % n)
                return
        optimizer.step(closure=optimizer_closure)



""" 

# Copyright 2019 by Gorka Munoz-Gil under the MIT license.
# This file is part of the Anomalous diffusion challenge (AnDi), and is
# released under the "MIT License Agreement". Please see the LICENSE
# file that should have been included in the repository containing the file
# (github.com/gorkamunoz/ANDI)

This file contains a recopilatory of diffusion models used in ANDI. The class
is organized in three subclasses, each corresponding to a different dimensionality
of the output trajectory.

Currently the library containts the following models:
    Function   Dimensions    Description
    - bm          (1D)       Brownian motion
    - fbm      (1D/2D/3D)    Fractional browian motion, simulated by the fbm python library
    - ctrw     (1D/2D/3D)    Continuous time random walks
    - lw       (1D/2D/3D)    Levy walks
    - attm     (1D/2D/3D)    Annealed transit time
    - sbm      (1D/2D/3D)    Scaled brownian motion
        
Inputs of generator functions:
    - T (int): lenght of the trajectory. Gets transformed to int if input
                is float.
    - alpha (float): anomalous exponent
    
    Some generator functions also have optional inputs, see each function for
    details.
                            
Outputs:
    - numpy.array of lenght d.T, where d is the dimension
    
    Some generator functions have optional outputs, see each function for details
"""

import numpy as np
import fbm
from .utils_andi import regularize, bm1D, sample_sphere
from math import pi as pi
from scipy.special import erfcinv
from functools import partial

# from .traj_tools import HiddenPrints
import warnings


__all__ = ["diffusion_models"]


def sym_exp():
    """
    exponential on both sides
    """
    return np.random.exponential() * np.random.choice([-1.0, 1.0])


def triangle():
    return np.random.triangular(-1, 0, 1)


class diffusion_models(object):
    def __init__(self):
        """Constructor of the class"""

    class oneD:
        """Class cointaning one dimensional diffusion models"""

        def fbm(self, T, alpha):
            """Creates a 1D fractional brownian motion trajectory"""
            H = alpha * 0.5
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                return np.reshape(fbm.fbm(int(T - 1), H), (-1, 1))

        def ctrw(self, T, alpha, regular_time=True):
            """Creates a 1D continuous time tandom walk trajectory
            Optional parameters:
                :regular_time (bool):
                    - True if to transform the trajectory to regular time."""
            if alpha > 1:
                raise ValueError(
                    "Continuous random walks only allow for anomalous exponents <= 1."
                )
            # Generate the waiting times from power-law distribution
            times = np.cumsum((1 - np.random.rand(T)) ** (-1 / alpha))
            times = times[: np.argmax(times > T) + 1]
            # Generate the positions of the walk
            positions = np.cumsum(np.random.randn(len(times)))
            positions -= positions[0]
            # Output
            if regular_time:
                return np.reshape(regularize(positions, times, T), (-1, 1))
            else:
                return np.stack((times, positions))

        def lw(self, T, alpha):
            """Creates a 1D Levy walk trajectory"""
            if alpha < 1:
                raise ValueError("Levy walks only allow for anomalous exponents > 1.")
            # Define exponents for the distribution of flight times
            if alpha == 2:
                sigma = np.random.rand()
            else:
                sigma = 3 - alpha
            dt = (1 - np.random.rand(T)) ** (-1 / sigma)
            dt[dt > T] = T + 1
            # Define the velocity
            # v = 10 * np.random.rand()
            v = 1
            # Generate the trajectory
            positions = np.empty(0)
            for t in dt:
                positions = np.append(
                    positions, v * np.ones(int(t)) * (2 * np.random.randint(0, 2) - 1)
                )
                if len(positions) > T:
                    break
            return np.reshape(np.cumsum(positions[: int(T)]) - positions[0], (-1, 1))

        def attm(self, T, alpha, regime=1):
            """Creates a 1D trajectory following the annealed transient time model
            Optional parameters:
                :regime (int):
                    - Defines the ATTM regime. Accepts three values: 0,1,2."""
            if regime not in [0, 1, 2]:
                raise ValueError("ATTM has only three regimes: 0, 1 or 2.")
            if alpha > 1:
                raise ValueError("ATTM only allows for anomalous exponents <= 1.")
            # Gamma and sigma selection
            if regime == 0:
                sigma = 3 * np.random.rand()
                gamma = np.random.uniform(low=-5, high=sigma)
                if alpha < 1:
                    raise ValueError(
                        "ATTM regime 0 only allows for anomalous exponents = 1."
                    )
            elif regime == 1:
                sigma = 3 * np.random.uniform(low=1e-2, high=1.1)
                gamma = sigma / alpha
                while sigma > gamma or gamma > sigma + 1:
                    sigma = 3 * np.random.uniform(low=1e-2, high=1.1)
                    gamma = sigma / alpha
            elif regime == 2:
                gamma = 1 / (1 - alpha)
                sigma = np.random.uniform(low=1e-2, high=gamma - 1)
            # Generate the trajectory
            positions = np.array([0])
            while len(positions) < T:
                Ds = (1 - np.random.uniform(low=0.1, high=0.99)) ** (1 / sigma)
                ts = Ds ** (-gamma)
                if ts > T:
                    ts = T
                positions = np.append(positions, positions[-1] + bm1D(ts, Ds))
            return np.reshape(positions[:T] - positions[0], (-1, 1))

        def sbm(self, T, alpha, sigma=1):
            """Creates a scaled brownian motion trajectory"""
            msd = (sigma**2) * np.arange(T + 1) ** alpha
            dx = np.sqrt(msd[1:] - msd[:-1])
            dx = np.sqrt(2) * dx * erfcinv(2 - 2 * np.random.rand(len(dx)))
            return np.reshape(np.cumsum(dx) - dx[0], (-1, 1))

        def ctrwExp(self, T, alpha, regular_time=True):
            """Creates a 1D continuous time tandom walk trajectory with exp step distribution
            Optional parameters:
                :regular_time (bool):
                    - True if to transform the trajectory to regular time."""
            if alpha > 1:
                raise ValueError(
                    "Continuous random walks only allow for anomalous exponents <= 1."
                )
            # Generate the waiting times from power-law distribution
            times = np.cumsum((1 - np.random.rand(T)) ** (-1 / alpha))
            times = times[: np.argmax(times > T) + 1]
            # Generate the positions of the walk
            shifts = [sym_exp() for _ in range(len(times))]
            positions = np.cumsum(shifts)
            positions -= positions[0]
            # Output
            if regular_time:
                return np.reshape(regularize(positions, times, T), (-1, 1))
            else:
                return np.stack((times, positions))

        def ctrwTri(self, T, alpha, regular_time=True):
            """Creates a 1D continuous time tandom walk trajectory with exp step distribution
            Optional parameters:
                :regular_time (bool):
                    - True if to transform the trajectory to regular time."""
            if alpha > 1:
                raise ValueError(
                    "Continuous random walks only allow for anomalous exponents <= 1."
                )
            # Generate the waiting times from power-law distribution
            times = np.cumsum((1 - np.random.rand(T)) ** (-1 / alpha))
            times = times[: np.argmax(times > T) + 1]
            # Generate the positions of the walk
            shifts = [triangle() for _ in range(len(times))]
            positions = np.cumsum(shifts)
            positions -= positions[0]
            # Output
            if regular_time:
                return np.reshape(regularize(positions, times, T), (-1, 1))
            else:
                return np.stack((times, positions))

        def BM(self, T, alpha, regular_time=True):
            """Creates a 1D Brownian Motion
            Optional parameters:
                :regular_time (bool):
                    - True if to transform the trajectory to regular time."""
            # Generate the waiting times from power-law distribution
            dt = 1.0
            dW = np.random.randn(T, 1) * np.sqrt(dt)
            X = np.cumsum(dW, axis=0)
            X -= X[0]
            return X

    class twoD(oneD):
        def ctrw(self, T, alpha, regular_time=True):
            """Creates a 2D continuous time tandom walk trajectory
            Optional parameters:
                :regular_time (bool):
                    - True if to transform the trajectory to regular time."""
            if alpha > 1:
                raise ValueError(
                    "Continuous random walks only allow for anomalous exponents <= 1."
                )
            # Generate the waiting times from power-law distribution
            # times = np.cumsum(np.random.rand(T) ** (-1 / alpha))
            times = np.cumsum((np.random.rand(T) ** (-1 / alpha)) - 1)
            times = times[: np.argmax(times > T) + 1]
            # Generate the positions of the walk
            posX = np.cumsum(np.random.randn(len(times)))
            posY = np.cumsum(np.random.randn(len(times)))
            posX -= posX[0]
            posY -= posY[0]
            # Regularize and output
            if regular_time:
                regX = regularize(posX, times, T)
                regY = regularize(posY, times, T)
                # replace concat by stack
                return np.stack((regX, regY), axis=1) / (1.0 - np.power(0.5, alpha))
            else:
                return np.stack((times, posX, posY))

        def fbm(self, T, alpha):
            """Creates a 2D fractional brownian motion trajectory"""
            # Defin Hurst exponent
            H = alpha * 0.5
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                return np.stack(
                    (
                        fbm.fbm(int(T - 1), H, length=int(T)),
                        fbm.fbm(int(T - 1), H, length=int(T)),
                    ),
                    axis=1,
                )

        def lw(self, T, alpha):
            """Creates a 2D Levy walk trajectory"""
            if alpha < 1:
                raise ValueError("Levy walks only allow for anomalous exponents > 1.")
            # Define exponents for the distribution of times
            if alpha == 2:
                sigma = np.random.rand()
            else:
                sigma = 3 - alpha
            dt = (1 - np.random.rand(T)) ** (-1 / sigma)
            dt[dt > T] = T + 1
            # Define the velocity
            # v = 10 * np.random.rand()
            v = 1.0
            # Define the array where we save step length
            d = np.empty(0)
            # Define the array where we save the angle of the step
            angles = np.empty(0)
            # Generate trajectory
            for t in dt:
                d = np.append(
                    d, v * np.ones(int(t)) * (2 * np.random.randint(0, 2) - 1)
                )
                angles = np.append(
                    angles, np.random.uniform(low=0, high=2 * pi) * np.ones(int(t))
                )
                if len(d) > T:
                    break
            d = d[: int(T)]
            angles = angles[: int(T)]
            posX, posY = [d * np.cos(angles), d * np.sin(angles)]
            return np.stack(
                (np.cumsum(posX) - posX[0], np.cumsum(posY) - posY[0]), axis=1
            )

        def attm(self, T, alpha, regime=1):
            """Creates a 2D trajectory following the annealed transient time model
            Optional parameters:
                :regime (int):
                    - Defines the ATTM regime. Accepts three values: 0,1,2."""
            if regime not in [0, 1, 2]:
                raise ValueError("ATTM has only three regimes: 0, 1 or 2.")
            if alpha > 1:
                raise ValueError("ATTM only allows for anomalous exponents <= 1.")
            # Gamma and sigma selection
            if regime == 0:
                sigma = 3 * np.random.rand()
                gamma = np.random.uniform(low=-5, high=sigma)
                if alpha < 1:
                    raise ValueError(
                        "ATTM regime 0 only allows for anomalous exponents = 1."
                    )
            elif regime == 1:
                sigma = 3 * np.random.uniform(low=1e-2, high=1.1)
                gamma = sigma / alpha
                while sigma > gamma or gamma > sigma + 1:
                    sigma = 3 * np.random.uniform(low=1e-2, high=1.1)
                    gamma = sigma / alpha
            elif regime == 2:
                gamma = 1 / (1 - alpha)
                sigma = np.random.uniform(low=1e-2, high=gamma - 1)
            # Generate the trajectory
            posX = np.array([0])
            posY = np.array([0])
            while len(posX) < T:
                Ds = (1 - np.random.uniform(low=0.1, high=0.99)) ** (1 / sigma)
                ts = Ds ** (-gamma)
                if ts > T:
                    ts = T
                posX = np.append(posX, posX[-1] + bm1D(ts, Ds))
                posY = np.append(posY, posY[-1] + bm1D(ts, Ds))
            return np.stack((posX[:T] - posX[0], posY[:T] - posY[0]), axis=1)

        def sbm(self, T, alpha, sigma=1):
            """Creates a scaled brownian motion trajectory"""
            msd = (sigma**2) * np.arange(T + 1) ** alpha
            deltas = np.sqrt(msd[1:] - msd[:-1])
            dx = np.sqrt(2) * deltas * erfcinv(2 - 2 * np.random.rand(len(deltas)))
            dy = np.sqrt(2) * deltas * erfcinv(2 - 2 * np.random.rand(len(deltas)))
            return np.stack((np.cumsum(dx) - dx[0], np.cumsum(dy) - dy[0]), axis=1)

        def ctrwExp(self, T, alpha, regular_time=True):
            """Creates a 2D continuous time tandom walk trajectory
            Optional parameters:
                :regular_time (bool):
                    - True if to transform the trajectory to regular time."""
            if alpha > 1:
                raise ValueError(
                    "Continuous random walks only allow for anomalous exponents <= 1."
                )
            # Generate the waiting times from power-law distribution
            times = np.cumsum((1 - np.random.rand(T)) ** (-1 / alpha))
            times = times[: np.argmax(times > T) + 1]
            # Generate the positions of the walk
            shifts_x = [sym_exp() for _ in range(len(times))]
            shifts_y = [sym_exp() for _ in range(len(times))]
            posX = np.cumsum(shifts_x)
            posY = np.cumsum(shifts_y)
            posX -= posX[0]
            posY -= posY[0]
            # Regularize and output
            if regular_time:
                regX = regularize(posX, times, T)
                regY = regularize(posY, times, T)
                return np.stack((regX, regY), axis=1)
            else:
                return np.stack((times, posX, posY))

        def ctrwTri(self, T, alpha, regular_time=True):
            """Creates a 2D continuous time tandom walk trajectory
            Optional parameters:
                :regular_time (bool):
                    - True if to transform the trajectory to regular time."""
            if alpha > 1:
                raise ValueError(
                    "Continuous random walks only allow for anomalous exponents <= 1."
                )
            # Generate the waiting times from power-law distribution
            times = np.cumsum((1 - np.random.rand(T)) ** (-1 / alpha))
            times = times[: np.argmax(times > T) + 1]
            # Generate the positions of the walk
            shifts_x = [triangle() for _ in range(len(times))]
            shifts_y = [triangle() for _ in range(len(times))]
            posX = np.cumsum(shifts_x)
            posY = np.cumsum(shifts_y)
            posX -= posX[0]
            posY -= posY[0]
            # Regularize and output
            if regular_time:
                regX = regularize(posX, times, T)
                regY = regularize(posY, times, T)
                return np.stack((regX, regY), axis=1)
            else:
                return np.stack((times, posX, posY))

        def BM(self, T, alpha, regular_time=True):
            """Creates a 2D Brownian Motion
            Optional parameters:
                :regular_time (bool):
                    - True if to transform the trajectory to regular time."""
            # Generate the waiting times from power-law distribution
            assert alpha == 1
            dt = 1.0
            dW = np.random.randn(T, 2) * np.sqrt(dt)
            X = np.cumsum(dW, axis=0)
            X -= X[0]
            return X

    class threeD(oneD):
        def ctrw(self, T, alpha, regular_time=True):
            """Creates a 3D continuous time tandom walk trajectory
            Optional parameters:
                :regular_time (bool):
                    - True if to transform the trajectory to regular time."""
            if alpha > 1:
                raise ValueError(
                    "Continuous random walks only allow for anomalous exponents <= 1."
                )
            # Generate the waiting times from power-law distribution
            times = np.cumsum((1 - np.random.rand(T)) ** (-1 / alpha))
            times = np.append(0, times)
            times = times[: np.argmax(times > T) + 1]
            # Generate the positions of the walk
            lengths = np.random.randn(len(times))
            posX, posY, posZ = np.cumsum(sample_sphere(len(times), lengths), axis=1)
            posX = posX - posX[0]
            posY = posY - posY[0]
            posZ = posZ - posZ[0]
            # Regularize and output
            if regular_time:
                regX = regularize(posX, times, T)
                regY = regularize(posY, times, T)
                regZ = regularize(posZ, times, T)
                return np.stack((regX, regY, regZ), axis=1)
            else:
                return np.stack((times, posX, posY, posZ))

        def fbm(self, T, alpha):
            """Creates a 3D fractional brownian motion trajectory"""
            # Define Hurst exponent
            H = alpha * 0.5
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                return np.stack(
                    (
                        fbm.fbm(int(T - 1), H),
                        fbm.fbm(int(T - 1), H),
                        fbm.fbm(int(T - 1), H),
                    ),
                    axis=1,
                )

        def lw(self, T, alpha, regular_time=True):
            """Creates a 3D Levy walk trajectory"""
            if alpha < 1:
                raise ValueError("Levy walks only allow for anomalous exponents > 1.")
            # Define exponents for the distribution of times
            if alpha == 2:
                sigma = np.random.rand()
            else:
                sigma = 3 - alpha
            dt = (1 - np.random.rand(T)) ** (-1 / sigma)
            dt[dt > T] = T + 1
            # Define the velocity
            # v = 10 * np.random.rand()
            v = 1.0
            # Create the trajectory
            posX = np.empty(0)
            posY = np.empty(0)
            posZ = np.empty(0)
            for t in dt:
                distX, distY, distZ = sample_sphere(1, v)
                posX = np.append(posX, distX * np.ones(int(t)))
                posY = np.append(posY, distY * np.ones(int(t)))
                posZ = np.append(posZ, distZ * np.ones(int(t)))
                if len(posX) > T:
                    break
            return np.stack(
                (
                    np.cumsum(posX[:T]) - posX[0],
                    np.cumsum(posY[:T]) - posY[0],
                    np.cumsum(posZ[:T]) - posZ[0],
                ),
                axis=1,
            )

        def attm(self, T, alpha, regime=1):
            """Creates a 3D trajectory following the annealed transient time model
            Optional parameters:
                :regime (int):
                    - Defines the ATTM regime. Accepts three values: 0,1,2."""
            if regime not in [0, 1, 2]:
                raise ValueError("ATTM has only three regimes: 0, 1 or 2.")
            if alpha > 1:
                raise ValueError("ATTM only allows for anomalous exponents <= 1.")
            # Parameter selection
            if regime == 0:
                sigma = 3 * np.random.rand()
                gamma = np.random.uniform(low=-5, high=sigma)
                if alpha < 1:
                    raise ValueError(
                        "ATTM Regime 0 can only produce trajectories with anomalous exponents = 1"
                    )
            elif regime == 1:
                sigma = 3 * np.random.uniform(low=1e-2, high=1.1)
                gamma = sigma / alpha
                while sigma > gamma or gamma > sigma + 1:
                    sigma = 3 * np.random.uniform(low=1e-2, high=1.1)
                    gamma = sigma / alpha
            elif regime == 2:
                gamma = 1 / (1 - alpha)
                sigma = np.random.uniform(low=1e-2, high=gamma - 1)
            # Create the trajectory
            posX = np.array([0])
            posY = np.array([0])
            posZ = np.array([0])
            while len(posX) < T:
                Ds = (1 - np.random.uniform(low=0.1, high=0.99)) ** (1 / sigma)
                ts = Ds ** (-gamma)
                if ts > T:
                    ts = T
                steps = np.sqrt(2 * Ds) * np.random.randn(int(ts))
                distX, distY, distZ = sample_sphere(len(steps), steps)
                posX = np.append(posX, posX[-1] + distX)
                posY = np.append(posY, posY[-1] + distY)
                posZ = np.append(posZ, posZ[-1] + distZ)
            return np.stack(
                (posX[:T] - posX[0], posY[:T] - posY[0], posZ[:T] - posZ[0]), axis=1
            )

        def sbm(self, T, alpha, sigma=1):
            """Creates a scaled brownian motion trajectory"""
            msd = (sigma**2) * np.arange(T + 1) ** alpha
            deltas = np.sqrt(msd[1:] - msd[:-1])
            dx = np.sqrt(2) * deltas * erfcinv(2 - 2 * np.random.rand(len(deltas)))
            dy = np.sqrt(2) * deltas * erfcinv(2 - 2 * np.random.rand(len(deltas)))
            dz = np.sqrt(2) * deltas * erfcinv(2 - 2 * np.random.rand(len(deltas)))
            return np.stack(
                (np.cumsum(dx) - dx[0], np.cumsum(dy) - dy[0], np.cumsum(dz) - dz[0]),
                axis=1,
            )

        def ctrwExp(self, T, alpha, regular_time=True):
            """Creates a 3D continuous time tandom walk trajectory
            Optional parameters:
                :regular_time (bool):
                    - True if to transform the trajectory to regular time."""
            # print("CTRW Exp")
            if alpha > 1:
                raise ValueError(
                    "Continuous random walks only allow for anomalous exponents <= 1."
                )
            # Generate the waiting times from power-law distribution
            times = np.cumsum((1 - np.random.rand(T)) ** (-1 / alpha))
            times = np.append(0, times)
            times = times[: np.argmax(times > T) + 1]
            # Generate the positions of the walk
            lengths = np.array([sym_exp() for _ in range(len(times))])
            posX, posY, posZ = np.cumsum(sample_sphere(len(times), lengths), axis=1)
            posX = posX - posX[0]
            posY = posY - posY[0]
            posZ = posZ - posZ[0]
            # Regularize and output
            if regular_time:
                regX = regularize(posX, times, T)
                regY = regularize(posY, times, T)
                regZ = regularize(posZ, times, T)
                return np.stack((regX, regY, regZ), axis=1)
            else:
                return np.stack((times, posX, posY, posZ))

        def ctrwTri(self, T, alpha, regular_time=True):
            """Creates a 3D continuous time tandom walk trajectory
            Optional parameters:
                :regular_time (bool):
                    - True if to transform the trajectory to regular time."""
            # print("CTRW Triangle")
            if alpha > 1:
                raise ValueError(
                    "Continuous random walks only allow for anomalous exponents <= 1."
                )
            # Generate the waiting times from power-law distribution
            times = np.cumsum((1 - np.random.rand(T)) ** (-1 / alpha))
            times = np.append(0, times)
            times = times[: np.argmax(times > T) + 1]
            # Generate the positions of the walk
            lengths = np.array([triangle() for _ in range(len(times))])
            posX, posY, posZ = np.cumsum(sample_sphere(len(times), lengths), axis=1)
            posX = posX - posX[0]
            posY = posY - posY[0]
            posZ = posZ - posZ[0]
            # Regularize and output
            if regular_time:
                regX = regularize(posX, times, T)
                regY = regularize(posY, times, T)
                regZ = regularize(posZ, times, T)
                return np.stack((regX, regY, regZ), axis=1)
            else:
                return np.stack((times, posX, posY, posZ))

        def BM(self, T, alpha, regular_time=True):
            # print("BM")
            """Creates a 3D Brownian Motion
            Optional parameters:
                :regular_time (bool):
                    - True if to transform the trajectory to regular time."""
            # Generate the waiting times from power-law distribution
            dt = 0.01
            dW = np.random.randn(T, 3) * np.sqrt(dt)
            X = np.cumsum(dW, axis=0)
            X -= X[0]
            return X


def generate_OU(D, T, log_theta, sigma):
    sub_sampling = int(max(1, np.power(10, 2 * log_theta + 1)))
    dt = 1.0 / sub_sampling
    dW = np.random.randn(sub_sampling * T, D) * np.sqrt(dt)
    theta = np.power(10, log_theta)
    X = np.zeros((T, D))
    x = X[0]
    for i in range(1, T * sub_sampling):
        x = x + theta * (-x) * dt + np.sqrt(2) * sigma * dW[i - 1]
        if i % sub_sampling == 0:
            X[i // sub_sampling] = x
            if i // sub_sampling == X.shape[0] - 1:
                break
    return X


generators = {
    1: {
        "ATTM": diffusion_models().oneD().attm,
        "CTRW": diffusion_models().oneD().ctrw,
        "fBM": diffusion_models().oneD().fbm,
        "LW": diffusion_models().oneD().lw,
        "sBM": diffusion_models().oneD().sbm,
        "BM": diffusion_models().oneD().BM,
        "OU": partial(generate_OU, D=1),
    },
    2: {
        "ATTM": diffusion_models().twoD().attm,
        "CTRW": diffusion_models().twoD().ctrw,
        "fBM": diffusion_models().twoD().fbm,
        "LW": diffusion_models().twoD().lw,
        "sBM": diffusion_models().twoD().sbm,
        "BM": diffusion_models().twoD().BM,
        "OU": partial(generate_OU, D=2),
    },
    3: {
        "ATTM": diffusion_models().threeD().attm,
        "CTRW": diffusion_models().threeD().ctrw,
        "fBM": diffusion_models().threeD().fbm,
        "LW": diffusion_models().threeD().lw,
        "sBM": diffusion_models().threeD().sbm,
        "BM": diffusion_models().threeD().BM,
        "OU": partial(generate_OU, D=3),
    },
}


def params_sampler(model, seed=0):
    if model == "CTRW":
        # return {"alpha": np.random.uniform(0.05, 1)}
        return {"alpha": np.random.uniform(0.1, 1.0)}
    elif model == "LW":
        return {"alpha": np.random.uniform(1.0, 2.0)}
    elif model == "fBM":
        return {"alpha": np.random.uniform(0.05, 1.95)}
    elif model == "fBM_fullrange":
        return {"alpha": np.random.uniform(0.01, 1.99)}
    elif model == "sBM":
        return {"alpha": np.random.uniform(0.1, 2.0)}
    elif model == "OU":
        # log(Theta) = 2.5
        # La courbe du MSD se sépare de celle du BM à 10 pas
        return {"log_theta": np.random.uniform(-2.0, 0.0), "sigma": 1.0}
    elif model == "ATTM":
        return {"alpha": np.random.uniform(0.05, 1.0)}
    elif model == "BM":
        return {"alpha": 1.0}
    else:
        raise NotImplementedError(f"Unknown model {model}")



from collections import defaultdict
import pytorch_lightning as pl
from torchmetrics import MeanAbsoluteError as MAE
from torchmetrics.classification.f_beta import F1Score as F1
import torch.nn as nn
import torch
from functools import partial
from ..layers.diverse import MLP, AlphaPredictor
from ..layers.features_init import *
from ..training.network_tools import L2_loss, Category_loss, is_concerned
from ..layers.encoders import *
from torch.optim.lr_scheduler import ExponentialLR
import scipy


class MainNet(pl.LightningModule):
    def __init__(
        self,
        tasks: list,
        n_c: int,  # number of convolutions
        latent_dim: int,
        dim: int,  # traj dim
        gamma: float = 0.98,
        lr: float = 1e-3,
        log_distance_features: bool = True,
        RW_types: list = ["fBM"],
        scale_types: list = ["step_std"],
    ):
        super().__init__()

        self.save_hyperparameters()
        x_dim = TrajsFeatures.x_dim(scale_types)
        e_dim = TrajsFeatures.e_dim(scale_types)
        self.save_hyperparameters({"x_dim": x_dim, "e_dim": e_dim})

        self.encoder = TrajsEncoder2(
            n_c=n_c,
            latent_dim=latent_dim,
            x_dim=self.hparams["x_dim"],
            e_dim=self.hparams["e_dim"],
            traj_dim=dim,
            n_scales=len(self.hparams["scale_types"])
            + 1,  # + 1 car on passe la longueur comme scale aussi
        )

        outputs = self.get_output_modules(tasks, latent_dim, dim, RW_types)

        self.features_maker = TrajsFeatures()

        self.out_networks = {}
        self.losses = {}
        self.targets = {}
        for out in outputs:
            net, target, loss = outputs[out]
            self.out_networks[out] = net
            self.targets[out] = target
            self.losses[out] = loss
        self.out_networks = nn.ModuleDict(self.out_networks)
        self.loss_scale = defaultdict(lambda: 1.0 / 12.0)
        self.set_loss_scale()

        if "alpha" in tasks:
            self.MAE = MAE()
        if "model" in tasks:
            self.F1 = F1(len(self.hparams["RW_types"]))

    def set_loss_scale(self):
        if "model" in self.losses:
            self.loss_scale["model"] = np.log(len(self.hparams["RW_types"]))
        if "alpha" in self.losses:
            self.loss_scale["alpha"] = scipy.stats.uniform.var(loc=0, scale=2)
        if "log_tau" in self.losses:
            self.loss_scale["log_tau"] = scipy.stats.uniform.var(
                loc=0, scale=np.log(50 / 5)
            )
        if "log_diffusion" in self.losses:
            self.loss_scale["log_diffusion"] = scipy.stats.uniform.var(loc=0, scale=8)

    def get_output_modules(self, tasks, latent_dim, dim, RW_types):
        outputs = {}
        if "alpha" in tasks:
            outputs["alpha"] = (
                AlphaPredictor(input_dim=latent_dim),
                partial(self.get_target, target="alpha"),
                # L2_loss,
                nn.MSELoss(),
            )
        if "model" in tasks:
            outputs["model"] = (
                MLP([latent_dim, 2 * latent_dim, latent_dim, len(RW_types)]),
                partial(self.get_target, target="model"),
                # Category_loss,
                nn.CrossEntropyLoss(),
            )
        if "log_theta" in tasks:
            outputs["log_theta"] = (
                MLP([latent_dim, 2 * latent_dim, latent_dim, 1]),
                partial(self.get_target, target="log_theta"),
                L2_loss,
            )
        if "log_tau" in tasks:
            outputs["log_tau"] = (
                MLP(
                    [latent_dim, 2 * latent_dim, latent_dim, latent_dim, 1],
                    out_range=(1, 2),
                ),
                partial(self.get_target, target="log_tau"),
                L2_loss,
            )
        if "log_diffusion" in tasks:
            outputs["log_diffusion"] = (
                MLP(
                    [latent_dim, 2 * latent_dim, latent_dim, latent_dim, 1],
                    out_range=(-4, 4),
                ),
                partial(self.get_target, target="log_diffusion"),
                L2_loss,
            )

        if len(outputs) == 0:
            print(tasks)
            raise BaseException("No outputs !")
        return outputs

    def get_target(self, data, target):
        if target == "alpha":
            return data.alpha
        elif target == "model":
            return data.model.view(
                -1,
            )
        elif target == "log_theta":
            return data.log_theta
        elif target == "log_tau":
            return data.log_tau
        elif target == "log_diffusion":
            return data.log_diffusion
        else:
            raise NotImplementedError("Unknown target %s" % target)

    def forward(self, x):

        X, E, scales, orientation = self.features_maker(
            x,
            scale_types=self.hparams["scale_types"],
            log_distance_features=self.hparams["log_distance_features"],
        )
        x.adj_t = x.adj_t.set_value(E, layout="coo")
        x.x = X
        x.scales = torch.cat([scales[k].view(-1, 1) for k in scales], dim=1)
        x.orientation = orientation

        assert x.x.shape[1] == self.hparams["x_dim"]
        h = self.encoder(x)
        out = {}

        for net in self.out_networks:
            out[net] = self.out_networks[net](h)

        return out, h

    def training_step(self, batch, batch_idx):
        loss, out, targets = self.shared_step(batch, stage="train")
        return {"loss": loss, "preds": out, "targets": targets}

    def test_step(self, batch, batch_idx):
        loss, _, _ = self.shared_step(batch, stage="test")
        return loss

    def on_test_epoch_end(self):
        self.logger.log_hyperparams(self.hparams)

    def validation_step(self, batch, batch_idx):
        loss, out, targets = self.shared_step(batch, stage="val")
        trajs_info = targets.copy()
        del targets["length"]
        return {
            "loss": loss,
            "preds": out,
            "targets": targets,
            "trajs_info": trajs_info,
        }

    def shared_step(self, batch, stage="train"):
        # Part of the computation that is common to train, val and test
        out, h = self(batch)
        targets = {}
        losses = {}
        weights = {}

        targets["length"] = batch.length

        for net in out:

            targets[net] = self.targets[net](batch)
            w = is_concerned(targets[net])
            weights[net] = torch.mean(1.0 * w)
            # print(f"{net} : <w> = {torch.mean(1.*w)}")
            # losses[net] = self.losses[net](out[net], targets[net], w)
            losses[net] = self.losses[net](out[net][w], targets[net][w])
            losses[net] /= self.loss_scale[net]
            self.log(
                "%s_%s_loss" % (net, stage),
                losses[net],
                on_step=False,
                on_epoch=True,
                logger=True,
                batch_size=h.shape[0],
            )

            if net == "alpha":
                self.MAE(out[net][w], targets[net][w])
                self.log(
                    "%s_MAE" % stage,
                    self.MAE,
                    on_step=False,
                    on_epoch=True,
                    prog_bar=stage == "val",
                    logger=True,
                    batch_size=h.shape[0],
                )
            elif net == "model":
                self.F1(out[net][w], targets[net][w])
                self.log(
                    "%s_F1" % stage,
                    self.F1,
                    on_step=False,
                    on_epoch=True,
                    prog_bar=stage == "val",
                    logger=True,
                    batch_size=h.shape[0],
                )

        out["latent"] = h
        # Pondération des loss en fonction du nombre de samples concernés
        loss = sum([losses[net] * weights[net] for net in losses])
        self.log(
            "%s_loss" % stage,
            loss,
            on_epoch=True,
            prog_bar=True,
            logger=True,
            batch_size=h.shape[0],
        )

        for key, value in out.items():
            out[key] = value.detach()

        return loss, out, targets

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(
            self.parameters(), amsgrad=True, lr=self.hparams["lr"]
        )
        scheduler = ExponentialLR(optimizer, gamma=self.hparams["gamma"])
        return [optimizer], [scheduler]



from typing import List, Union
import torch.nn as nn
from torch_geometric.data import Batch
from torch_scatter import scatter
import torch


def scatter_std(X, B):
    avg_X_2 = scatter(src=X**2, index=B, dim=0, reduce="mean")
    avg_X = scatter(src=X, index=B, dim=0, reduce="mean")
    Var = avg_X_2 - (avg_X**2)
    return torch.sqrt(Var)


def is_first_point(B):
    # print(B)
    f = torch.abs(B - torch.roll(B, 1, dims=(0,))) > 0
    # print("first points", f)
    if torch.max(B) > torch.min(B):
        assert f[0] == True
    return f


def is_last_point(B):
    # print(B)
    f = torch.abs(B - torch.roll(B, -1, dims=(0,))) > 0
    # print("last points", f)
    if torch.max(B) > torch.min(B):
        assert f[-1] == True
    return f


def diff_per_graph(X, B, fill_last=True):
    d = torch.roll(X, -1, dims=(0,)) - X
    if fill_last:
        # d[is_last_point(B)] = 0.0
        return torch.where(
            torch.stack([is_last_point(B) for d in range(X.shape[1])], dim=1),
            0.0,
            d.double(),
        ).float()
    else:
        d = d[~is_last_point(B)]
        return d, B[~is_last_point(B)]


def cumsum_per_graph(X, B):
    cumsum = torch.cumsum(X, dim=0)
    min_per_b = scatter(cumsum, B, dim=0, reduce="min")

    first_point = 1 * is_first_point(B)
    first_value = scatter(X * first_point, B, dim=0, reduce="sum")

    cumsum = cumsum - torch.index_select(input=min_per_b, dim=0, index=B)
    cumsum = cumsum + torch.index_select(input=first_value, dim=0, index=B)
    return cumsum


def cummax_per_graph(X, B):
    cummax, _ = torch.cummax(X, dim=0)
    min_per_b = scatter(cummax, B, dim=0, reduce="min")

    first_point = 1 * is_first_point(B)
    first_value = scatter(X * first_point, B, dim=0, reduce="sum")

    cumsum = cummax - torch.index_select(input=min_per_b, dim=0, index=B)
    cumsum = cummax + torch.index_select(input=first_value, dim=0, index=B)
    return cumsum


class TrajsFeatures(nn.Module):
    def get_scales(self, B, P, scale_types):
        scales = {}

        if scale_types != ["pos_std"]:
            dr = diff_per_graph(P, B)
            in_points = ~is_first_point(B)
            dr = dr[in_points]
            dr_norm = torch.sqrt(1e-5 + torch.sum(dr**2, dim=1))

        if "pos_std" in scale_types:
            P_STD = torch.sqrt(torch.sum(scatter_std(P, B) ** 2, dim=1))
            scales["pos_std"] = P_STD

        if "step_sum" in scale_types:
            step_sum = scatter(dr_norm, index=B[in_points], dim=0, reduce="sum")
            scales["step_sum"] = step_sum

        if "step_std" in scale_types:
            step_std = scatter_std(dr_norm, B[in_points])
            scales["step_std"] = step_std

        if "step_mean" in scale_types:
            step_mean = scatter(dr_norm, index=B[in_points], dim=0, reduce="mean")
            scales["step_mean"] = step_mean

        if "step_var" in scale_types:
            # var = <dr^2> - <dr>^2
            step_var = scatter(dr_norm**2, index=B[in_points], dim=0, reduce="mean")
            # - torch.sum(
            #    scatter(dr, index=B[in_points], dim=0, reduce="mean") ** 2, dim=1
            # )
            scales["step_var"] = step_var
        return scales

    @classmethod
    def x_dim(cls, scale_types):
        return 1 + len(scale_types) * 4

    @classmethod
    def e_dim(cls, scale_types):
        return 2 + len(scale_types) * 5

    def forward(
        self,
        data: Batch,
        scale_types: List[str] = ["pos_std"],
        return_intermediate: bool = False,
        log_distance_features: bool = False,
    ):

        B = data.batch
        P = data.pos

        row, col, _ = data.adj_t.t().coo()
        scales = self.get_scales(B, P, scale_types)

        time = cumsum_per_graph(torch.ones_like(B), B)
        L = scatter(torch.ones_like(B), B, dim=0, reduce="sum")
        time_norm = time / torch.index_select(L, dim=0, index=B)

        node_features = [time_norm.view(-1, 1).float()]
        edge_features = [(time_norm[row] - time_norm[col]).view(-1, 1).float()]

        dr = diff_per_graph(P, B)

        norm = lambda x: torch.log(1e-8 + x) if log_distance_features else x

        for i, (scale_name, scale) in enumerate(scales.items()):
            scale_factor = torch.index_select(scale, dim=0, index=B).view(-1, 1)
            dr_ = dr / scale_factor
            p = P / scale_factor
            dr_norm = torch.sqrt(1e-5 + torch.sum(dr_**2, dim=1))

            cum_dist = cumsum_per_graph(dr_norm, B).view(-1, 1)
            cum_msd = cumsum_per_graph(dr_norm**2, B).view(-1, 1)
            node_features.append(norm(cum_dist))
            node_features.append(norm(cum_msd))
            node_features.append(norm(dr_norm.view(-1, 1)))
            node_features.append(norm(cummax_per_graph(dr_norm, B).view(-1, 1)))

            end, start = p[row], p[col]
            d = end - start
            d = torch.sqrt(torch.sum(d**2, dim=1))

            end_jump, start_jump = dr_[row], dr_[col]
            corr = torch.sum(end_jump * start_jump, dim=1) / (
                torch.linalg.norm(end_jump, dim=1)
                * torch.linalg.norm(start_jump, dim=1)
            )
            corr[torch.isnan(corr)] = 0.0
            corr[torch.isinf(corr)] = 0.0

            edge_features.append(norm(d.view(-1, 1)))
            if i == 0:
                edge_features.append(corr.view(-1, 1))
            edge_features.append(torch.sign(cum_dist[row] - cum_dist[col]))
            edge_features.append(norm(torch.abs(cum_dist[row] - cum_dist[col])))
            edge_features.append(torch.sign(cum_msd[row] - cum_msd[col]))
            edge_features.append(norm(torch.abs(cum_msd[row] - cum_msd[col])))

        X = torch.cat(node_features, dim=1)
        E = torch.cat(edge_features, dim=1)

        assert X.shape[1] == TrajsFeatures.x_dim(scale_types)
        assert E.shape[1] == TrajsFeatures.e_dim(scale_types)
        assert torch.sum(torch.isnan(X)) == 0
        assert torch.sum(torch.isnan(E)) == 0, torch.sum(torch.isnan(E), dim=0)
        assert torch.sum(torch.isinf(X)) == 0
        assert torch.sum(torch.isinf(E)) == 0

        # Make a tensor with scales and L (to be used to infer D)
        u = scatter(dr, B, reduce="mean", dim=0)
        u = u / torch.sqrt(1e-5 + torch.sum(u**2, dim=1).view(-1, 1))
        scales["log_L"] = torch.log(L.float())

        orientation = u

        if not return_intermediate:
            return X, E, scales, orientation
        else:
            return X, E, scales, orientation, node_features, edge_features, dr_norm



from .skeleton_data import *


def TrajData(
    raw_positions, graph_info={}, traj_info={}, original_positions=None
) -> SkeletonTrajData:

    assert raw_positions is not None, "raw_positions is None"
    if graph_info["data_type"] == "invariant":
        return InvariantTrajData(
            raw_positions,
            graph_info=graph_info,
            traj_info=traj_info,
            original_positions=original_positions,
        )

    elif graph_info["data_type"] == "node_features":
        return NodeFeaturesTrajData(
            raw_positions,
            graph_info=graph_info,
            traj_info=traj_info,
            original_positions=original_positions,
        )

    elif graph_info["data_type"] == "no_features":
        return SkeletonTrajData(
            raw_positions,
            graph_info=graph_info,
            traj_info=traj_info,
            original_positions=original_positions,
        )

    raise BaseException("Unknown data type : %s" % graph_info["data_type"])


class NodeFeaturesTrajData(SkeletonTrajData):
    def get_node_features(self, positions, graph_info) -> np.array:
        features = []
        N = positions.shape[0]
        reshape = lambda a: np.reshape(a, (N - 1, -1))
        norm = lambda x: x
        if graph_info["log_features"]:
            norm = lambda x: np.log(1e-5 + x)
        # Time
        norm_time = np.arange(N - 1) / (N - 1)
        features.append(reshape(norm_time))
        # get displacements
        dr = get_steps(positions)
        dr_vec = positions[1:] - positions[:-1]

        MD = reshape(np.cumsum(dr) / (1.0 + np.arange(N - 1)))
        MSD = reshape(np.power(np.cumsum(dr ** 2), 1.0 / 2) / (1.0 + np.arange(N - 1)))
        MQD = reshape(np.power(np.cumsum(dr ** 4), 1.0 / 4) / (1.0 + np.arange(N - 1)))
        MaxD = reshape(np.maximum.accumulate(dr))

        for scale_name in graph_info["scale_types"]:
            scale = traj_scale(positions, scale_name)
            assert scale > 0.0, "scale not positive %s" % positions

            features.append(np.clip(positions[:-1] / scale, -10, 10))
            features.append(np.clip(dr_vec / scale, -10, 10))

            features.append(norm(MD / scale))
            features.append(norm(MSD / scale))
            features.append(norm(MQD / scale))
            features.append(norm(MaxD / scale))
        features = np.concatenate(features, axis=1)
        assert np.sum(np.isnan(features)) == 0
        assert np.sum(np.isinf(features)) == 0
        return features


class InvariantTrajData(SkeletonTrajData):
    # Invariant by rotation

    def get_edge_features(self, X, positions, edge_index, graph_info):
        L = positions.shape[0]
        E = edge_index.shape[1]
        dt = (edge_index[1] - edge_index[0]) / L

        reshape = lambda a: np.reshape(a, (E, -1))
        DR = np.sqrt(
            np.sum((positions[edge_index[0]] - positions[edge_index[1]]) ** 2, axis=1)
        )
        features = [reshape(dt)]
        for scale_name in graph_info["scale_types"]:
            scale = traj_scale(positions, scale_name)
            features.append(reshape(DR / scale))
        features = np.concatenate(features, axis=1)
        assert np.sum(np.isnan(features)) == 0
        assert np.sum(np.isinf(features)) == 0
        return features

    def get_node_features(self, positions, graph_info) -> np.array:
        features = []
        N = positions.shape[0]
        reshape = lambda a: np.reshape(a, (N - 1, -1))
        norm = lambda x: x
        if graph_info["log_features"]:
            norm = lambda x: np.log(1e-5 + x)
        # Time
        norm_time = np.arange(N - 1) / (N - 1)
        features.append(reshape(norm_time))
        # get displacements
        dr = get_steps(positions)

        MD = reshape(np.cumsum(dr) / (1.0 + np.arange(N - 1)))
        MSD = reshape(np.power(np.cumsum(dr ** 2), 1.0 / 2) / (1.0 + np.arange(N - 1)))
        MQD = reshape(np.power(np.cumsum(dr ** 4), 1.0 / 4) / (1.0 + np.arange(N - 1)))
        MaxD = reshape(np.maximum.accumulate(dr))

        for scale_name in graph_info["scale_types"]:
            scale = traj_scale(positions, scale_name)
            assert scale > 0.0, "scale not positive %s" % positions
            features.append(norm(MD / scale))
            features.append(norm(MSD / scale))
            features.append(norm(MQD / scale))
            features.append(norm(MaxD / scale))
        features = np.concatenate(features, axis=1)
        assert np.sum(np.isnan(features)) == 0
        assert np.sum(np.isinf(features)) == 0
        return features



from collections import defaultdict
import pytorch_lightning as pl
from torchmetrics import MeanAbsoluteError as MAE
from torchmetrics.classification.f_beta import F1Score as F1
import torch.nn as nn
import torch
from functools import partial
from ..layers.diverse import MLP, AlphaPredictor
from ..layers.features_init import *
from ..training.network_tools import L2_loss, Category_loss, is_concerned
from ..layers.encoders import *
from torch.optim.lr_scheduler import ExponentialLR
import scipy


class MainNet(pl.LightningModule):
    def __init__(
        self,
        n_c: int,  # number of convolutions
        latent_dim: int,
        dim: int,  # traj dim
        gamma: float = 0.98,
        lr: float = 1e-3,
        log_distance_features: bool = True,
        RW_types: list = ["fBM"],
        scale_types: list = ["step_std"],
    ):
        super().__init__()

        self.save_hyperparameters()
        x_dim = TrajsFeatures.x_dim(scale_types)
        e_dim = TrajsFeatures.e_dim(scale_types)
        self.save_hyperparameters({"x_dim": x_dim, "e_dim": e_dim})

        self.encoder = TrajsEncoder2(
            n_c=n_c,
            latent_dim=latent_dim,
            x_dim=self.hparams["x_dim"],
            e_dim=self.hparams["e_dim"],
            traj_dim=dim,
            n_scales=len(self.hparams["scale_types"])
            + 1,  # + 1 car on passe la longueur comme scale aussi
        )

        outputs = {}
        outputs["alpha"] = (
            AlphaPredictor(input_dim=latent_dim),
            partial(self.get_target, target="alpha"),
            # L2_loss,
            nn.MSELoss(),
        )
        outputs["model"] = (
            MLP([latent_dim, 2 * latent_dim, latent_dim, len(RW_types)]),
            partial(self.get_target, target="model"),
            # Category_loss,
            nn.CrossEntropyLoss(),
        )

        self.features_maker = TrajsFeatures()

        self.out_networks = {}
        self.losses = {}
        self.targets = {}
        for out in outputs:
            net, target, loss = outputs[out]
            self.out_networks[out] = net
            self.targets[out] = target
            self.losses[out] = loss
        self.losses = nn.ModuleDict(self.losses)
        self.out_networks = nn.ModuleDict(self.out_networks)
        self.loss_scale = defaultdict(lambda: 1.0 / 12.0)
        self.set_loss_scale()

        self.MAE_train = MAE()
        self.MAE_val = MAE()
        self.MAE_test = MAE()
        self.F1_train = F1(len(self.hparams["RW_types"]))
        self.F1_val = F1(len(self.hparams["RW_types"]))
        self.F1_test = F1(len(self.hparams["RW_types"]))
        self.metrics = {}
        self.metrics[("alpha", "train")] = self.MAE_train
        self.metrics[("model", "train")] = self.F1_train
        self.metrics[("alpha", "val")] = self.MAE_val
        self.metrics[("model", "val")] = self.F1_val
        self.metrics[("alpha", "test")] = self.MAE_test
        self.metrics[("model", "test")] = self.F1_test

    def set_loss_scale(self):
        if "model" in self.losses:
            self.loss_scale["model"] = np.log(len(self.hparams["RW_types"]))
        if "alpha" in self.losses:
            self.loss_scale["alpha"] = scipy.stats.uniform.var(loc=0, scale=2)

    def get_target(self, data, target):
        if target == "alpha":
            return data.alpha
        elif target == "model":
            return data.model.view(
                -1,
            )
        else:
            raise NotImplementedError("Unknown target %s" % target)

    def forward(self, x):

        X, E, scales, orientation = self.features_maker(
            x,
            scale_types=self.hparams["scale_types"],
            log_distance_features=self.hparams["log_distance_features"],
        )
        x.adj_t = x.adj_t.set_value(E, layout="coo")
        x.x = X
        x.scales = torch.cat([scales[k].view(-1, 1) for k in scales], dim=1)
        x.orientation = orientation

        assert x.x.shape[1] == self.hparams["x_dim"]
        h = self.encoder(x)
        out = {}

        for net in self.out_networks:
            out[net] = self.out_networks[net](h)

        return out, h

    # def on_fit_start(self):
    #    print("Self.device")
    #    print(self.device)
    #    for key, v in self.metrics.items():
    #        v = v.to(self.device)

    def training_step(self, batch, batch_idx):
        loss, out, targets = self.shared_step(batch, stage="train")
        return loss
        # return {"loss": loss, "preds": out, "targets": targets}

    def test_step(self, batch, batch_idx):
        loss, _, _ = self.shared_step(batch, stage="test")
        return loss

    def on_test_epoch_end(self):
        self.logger.log_hyperparams(self.hparams)

    def validation_step(self, batch, batch_idx):
        loss, out, targets = self.shared_step(batch, stage="val")
        trajs_info = targets.copy()
        del targets["length"]
        return {
            "loss": loss,
            "preds": out,
            "targets": targets,
            "trajs_info": trajs_info,
        }

    def shared_step(self, batch, stage="train"):
        # Part of the computation that is common to train, val and test
        out, h = self(batch)
        targets = {}
        losses = {}
        weights = {}

        targets["length"] = batch.length

        for net in out:

            targets[net] = self.targets[net](batch)
            w = is_concerned(targets[net])
            weights[net] = torch.mean(1.0 * w)
            # print(f"{net} : <w> = {torch.mean(1.*w)}")
            # losses[net] = self.losses[net](out[net], targets[net], w)
            losses[net] = self.losses[net](out[net][w], targets[net][w])
            losses[net] /= self.loss_scale[net]
            self.log(
                "%s_%s_loss" % (net, stage),
                losses[net],
                on_step=False,
                on_epoch=True,
                logger=True,
                batch_size=h.shape[0],
            )

            self.metrics[(net, stage)](out[net][w], targets[net][w])
            self.log(
                "%s_%s_%s"
                % (net, stage, self.metrics[(net, stage)].__class__.__qualname__),
                self.metrics[(net, stage)],
                on_step=False,
                on_epoch=True,
                prog_bar=stage != "val",
                logger=True,
                batch_size=h.shape[0],
            )

        out["latent"] = h
        # Pondération des loss en fonction du nombre de samples concernés
        loss = sum([losses[net] * weights[net] for net in losses])
        self.log(
            "%s_loss" % stage,
            loss,
            on_epoch=True,
            prog_bar=True,
            logger=True,
            batch_size=h.shape[0],
        )

        for key, value in out.items():
            out[key] = value.detach()

        return loss, out, targets

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(
            self.parameters(), amsgrad=True, lr=self.hparams["lr"]
        )
        scheduler = ExponentialLR(optimizer, gamma=self.hparams["gamma"])
        return [optimizer], [scheduler]
